# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
# flake8: noqa
"""
A copy of sample_experiment.py with multiple sessions.
To produce the data directory for this experiment use (make sure
that the directory you specify exists)::

    >>> from eelbrain import datasets
    >>> datasets.setup_samples_experiment('~/Data', n_segments=2, n_sessions=2, name='SampleExperimentSessions')

Then you can use::

    >>> from sample_experiment_sessions import SampleExperiment, ROOT
    >>> e = SampleExperiment(ROOT)

"""
from eelbrain.pipeline import *


ROOT = "~/Data/SampleExperimentSessions"


class SampleExperiment(MneExperiment):

    meg_system = 'neuromag306mag'
    stim_channel = 'STI 014'

    raw = {
        '0-40': RawFilter('raw', None, 40, method='iir'),
        '1-40': RawFilter('raw', 1, 40, method='iir'),
    }

    variables = {
        'event': {(1, 2, 3, 4): 'target', 5: 'smiley', 32: 'button'},
        'side': {(1, 3): 'left', (2, 4): 'right'},
        'modality': {(1, 2): 'auditory', (3, 4): 'visual'}
    }

    epochs = {
        'target1': PrimaryEpoch('sample1', "event == 'target'", decim=5),
        'target2': PrimaryEpoch('sample2', "event == 'target'", decim=5),
        'super': SuperEpoch(('target1', 'target2')),
    }



"Combine icons into a Python file"
import os
from wx.tools.img2py import img2py

python_file = "../eelbrain/_wxgui/icons.py"

# individual files
files = {}  # name: file

# folders that are included entirely
folders = [
    "actions",
    "brain",
    "copy",
    "documents",
    "plot",
]

# tango files (in `tango` subfolder)
tango = {
    'actions': [
        'document-open',
        'document-save',
        'document-save-as',
        'edit-clear',
        'edit-redo',
        'edit-undo',
        'go-down',
        'go-first',
        'go-last',
        'go-next',
        'go-previous',
        'go-up',
        'media-seek-forward',
        'media-seek-backward',
        'system-log-out',
        'view-refresh',
    ],
    'apps': [
        'help-browser',
        'utilities-terminal'
    ],
    'mimetypes': ['x-office-presentation'],
    'places': ['start-here'],
    'status': ['image-missing'],
}


for folder in folders:
    for name in os.listdir(folder):
        if name.endswith('.png'):
            path = os.path.join(folder, name)
            iconname = '/'.join((folder, name[:-4]))
            files[iconname] = path

for basename, names in tango.iteritems():
    for name in names:
        path = os.path.join('tango', basename, name + '.png')
        iconname = '/'.join(('tango', basename, name))
        files[iconname] = path


kwargs = dict(compressed=True,
              catalog=True,
              functionCompatible=False,
              functionCompatibile=-1)


img2py('system-icons/eelbrain/eelbrain256.png', python_file, append=False,
       imgName='eelbrain256', icon=False, **kwargs)

img2py('system-icons/eelbrain/eelbrain32.png', python_file, append=True,
       imgName='eelbrain', icon=True, **kwargs)

for name, image_file in files.iteritems():
    img2py(image_file, python_file, append=True,
           imgName=name, icon=False, **kwargs)

print("Done")



from eelbrain import *

from eelbrain._wxgui import get_app

if __name__ == '__main__':
    get_app()
    ds = datasets.get_uts(True, nrm=True)
    # res = testnd.TTestRelated('utsnd', 'A', match='rm', ds=ds, samples=100)
    res2 = testnd.ANOVA('utsnd', 'A*B*nrm(A)', data=ds, samples=100, pmin=0.5)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
import numpy as np
from numpy import newaxis
from scipy.ndimage import generate_binary_structure

from .._data_obj import Dimension


class Adjacency:
    """N-dimensional adjacency"""
    __slots__ = ('struct', 'custom', 'vector')

    def __init__(self, dims, parc=None):
        types = [dim._adjacency_type for dim in dims]
        invalid = set(types).difference(Dimension._ADJACENCY_TYPES)
        if invalid:
            raise RuntimeError(f"Invalid adjacency type: {', '.join(invalid)}")

        # vector: will be collapsed by test function
        n_vector = types.count('vector')
        if n_vector > 1:
            raise NotImplementedError("More than one axis with vector adjacency")
        elif n_vector:
            self.vector = types.index('vector')
            types = types[:self.vector] + type[self.vector + 1:]
        else:
            self.vector = None

        # custom adjacency
        self.custom = {}
        n_custom = types.count('custom')
        if n_custom > 1:
            raise NotImplementedError("More than one axis with custom adjacency")
        elif n_custom:
            axis = types.index('custom')
            if axis > 0:
                raise NotImplementedError(
                    "Custom adjacency on axis other than first")
            custom_dim = dims[axis]
            if custom_dim.name == parc:
                edges = custom_dim.adjacency(disconnect_parc=True)
            else:
                edges = custom_dim.adjacency()
            dim_length = len(custom_dim)
            src = edges[:, 0]
            n_edges = np.bincount(src, minlength=dim_length)
            edge_stop = np.cumsum(n_edges)
            edge_start = edge_stop - n_edges
            self.custom[axis] = (edges, edge_start, edge_stop)

        # prepare struct for grid adjacency
        self.struct = generate_binary_structure(len(dims), 1)
        for i, ctype in enumerate(types):
            if ctype != 'grid':
                self.struct[(slice(None),) * i + (slice(None, None, 2),)] = False

    def __getstate__(self):
        return {k: getattr(self, k) for k in self.__slots__}

    def __setstate__(self, state):
        for k, v in state.items():
            setattr(self, k, v)


def find_peaks(x, adjacency, out=None):
    """Find peaks (local maxima, including plateaus) in x

    Returns
    -------
    out : array (x.shape, bool)
        Boolean array which is True only on local maxima.
    """
    if out is None:
        out = np.empty(x.shape, bool)
    elif out.dtype.kind != 'b':
        raise TypeError(f"{out.dtype=}, needs to be array of boolean type")
    elif out.shape != x.shape:
        raise ValueError(f"{out.shape=}, needs to be same as {x.shape=}")
    out.fill(True)

    # move through each axis in both directions and discard descending
    # slope. Do most computationally intensive axis last.
    for ax in range(x.ndim - 1, -1, -1):
        if ax in adjacency.custom:
            shape = (len(x), -1)
            xsa = x.reshape(shape)
            outsa = out.reshape(shape)
            axlen = xsa.shape[1]

            conn_src, conn_dst = adjacency.custom[ax][0].T
            for i in range(axlen):
                data = xsa[:, i]
                outslice = outsa[:, i]
                if not np.any(outslice):
                    continue

                # find all points under a slope
                sign = np.sign(data[conn_src] - data[conn_dst])
                no = set(conn_src[sign < 0])
                no.update(conn_dst[sign > 0])

                # expand to equal points
                border = no
                while border:
                    # forward
                    idx = np.isin(conn_src, border)
                    conn_dst_sub = conn_dst[idx]
                    eq = np.equal(data[conn_src[idx]], data[conn_dst_sub])
                    new = set(conn_dst_sub[eq])
                    # backward
                    idx = np.isin(conn_dst, border)
                    conn_src_sub = conn_src[idx]
                    eq = np.equal(data[conn_src_sub], data[conn_dst[idx]])
                    new.update(conn_src_sub[eq])

                    # update
                    new.difference_update(no)
                    no.update(new)
                    border = new

                # mark vertices or whole isoline
                if no:
                    outslice[list(no)] = False
                elif not np.all(outslice):
                    outslice.fill(False)
        else:
            if x.ndim == 1:
                xsa = x[:, newaxis]
                outsa = out[:, newaxis]
            else:
                xsa = x.swapaxes(0, ax)
                outsa = out.swapaxes(0, ax)
            axlen = len(xsa)

            kernel = np.empty(xsa.shape[1:], bool)

            diff = np.diff(xsa, 1, 0)

            # forward
            kernel.fill(True)
            for i in range(axlen - 1):
                kernel &= outsa[i]
                kernel[diff[i] > 0] = True
                kernel[diff[i] < 0] = False
                outsa[i + 1] *= kernel

            # backward
            kernel.fill(True)
            for i in range(axlen - 2, -1, -1):
                kernel &= outsa[i + 1]
                kernel[diff[i] < 0] = True
                kernel[diff[i] > 0] = False
                outsa[i] *= kernel

    return out



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""
.. _exa-impulse:

Impulse predictors for epochs
=============================
:func:`epoch_impulse_predictor` generates predictor variables for reverse
correlation in trial-based experiments with discrete events. The function
generates one impulse per trial, and these impulsescan be of varaiable magnitude
and have variable latency.

The example uses simulated data meant to vaguely resemble data from an N400
experiment, but not intended as a physiologically realistic simulation.

.. contents:: Sections
    :local:
    :backlinks: top

"""
# sphinx_gallery_thumbnail_number = 2
from eelbrain import *

ds = datasets.simulate_erp(snr=1)
print(ds.summary())

###############################################################################
# Discrete events
# ---------------
# Computing a TRF for an impulse at trial onset is very similar to averaging:

any_trial = epoch_impulse_predictor('eeg', 1, data=ds)
fit = boosting('eeg', any_trial, -0.100, 0.600, basis=0.050, data=ds, partitions=2, delta=0.01)
average = ds['eeg'].mean('case')
trf = fit.h.sub(time=(average.time.tmin, average.time.tstop))
p = plot.TopoButterfly([fit.h_scaled, average], xlim=(-0.100, 0.600), axtitle=['Impulse response', 'Average'], t=0.400)

###############################################################################
# Categorial variables
# --------------------
# Impulse predictors can be used like dummy codes in a regression model.
# Use one impulse to code for occurrence of any word (``any_word``), and a
# second impulse to code for unpredictable words only (``cloze``):

any_word = epoch_impulse_predictor('eeg', 1, data=ds, name='any_word')
# effect code for cloze (1 for low cloze, -1 for high cloze)
cloze_code = Var.from_dict(ds['predictability'], {'high': 0, 'low': 1})
low_cloze = epoch_impulse_predictor('eeg', cloze_code, data=ds, name='low_cloze')

# plot the predictors for each trial
p = plot.UTS([any_word, low_cloze], '.case', stem=True)

###############################################################################
# Estimate response functions for these two predictors. Based on the coding,
# ``any_word`` reflects the response to predictable words, and ``low_cloze``
# reflects how unpredictable words differ from predictable words:

fit = boosting('eeg', [any_word, low_cloze], 0, 0.5, basis=0.050, model='predictability', data=ds, partitions=2, delta=0.01)
p = plot.TopoButterfly(fit.h, xlim=(-0.100, 0.600), t=0.400)

###############################################################################
# Continuous variables
# --------------------
# Impulse predictors can similarly accommodate continuous variables:

# effect code for cloze (1 for low cloze, -1 for high cloze)
n_chars = epoch_impulse_predictor('eeg', 'n_chars', data=ds, name='n_chars')
surprisal = epoch_impulse_predictor('eeg', '-numpy.log2(cloze)', data=ds, name='surprisal')

# plot the predictors for each trial
p = plot.UTS([n_chars, surprisal, any_word], '.case', stem=True)

###############################################################################
# Estimate response functions. Based on the coding, ``any_word`` reflects the
# hypothetical response to a words with 0 characters and 0 surprisal, whereas
# the two other predictors reflect the change in response by ``n_chars`` and
# surprisal:

fit = boosting('eeg', [any_word, n_chars, surprisal], 0, 0.5, basis=0.050, model='predictability', data=ds, partitions=2, delta=0.01)
p = plot.TopoButterfly(fit.h, xlim=(-0.100, 0.600), t=0.400)




# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"A few basic operations needed throughout Eelbrain"
from collections import defaultdict
from dataclasses import dataclass, fields
import functools
import inspect
import logging
import re
from typing import Dict
from warnings import warn

from .notebooks import tqdm


LOG_LEVELS = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'ERROR': logging.ERROR,
    'CRITICAL': logging.CRITICAL,
}


def as_list(item):
    if isinstance(item, list):
        return item
    else:
        return [item]


def as_sequence(items, item_type=str):
    if isinstance(items, item_type):
        return items,
    return items


def ask(
        message: str,
        options: Dict[str, str],
        allow_empty: bool = False,
        help: str = None,
        default: str = '',
) -> str:
    """Ask user for input

    Parameters
    ----------
    message
        Message.
    options
        ``{command: description}`` mapping.
    allow_empty
        Allow empty string as command.
    help
        If provided, add a "help" option that prints ``help``.
    default
        Default answer; implies ``allow_empty``, but will substitute ``default``
        when user input is the empty string.

    Returns
    -------
    command : str
        The user command.
    """
    options = dict(options)
    if help is not None:
        assert 'help' not in options
        options['help'] = 'display help'
    if default:
        options[default] += ' (default)'
        allow_empty = True
    print(message)
    print('---')
    print('\n'.join(f'{k}:  {v}' for k, v in options.items()))
    while True:
        command = input(" > ") or default
        if command in options or (allow_empty and not command):
            if help is not None and command == 'help':
                print(help)
            else:
                return command
        else:
            print(f"Invalid entry - type one of ({', '.join(options)})")


def deprecated(version, replacement):
    """Decorator to deprecate functions and methods

    Also handles docstring of the deprecated function.

    Parameters
    ----------
    version : str
        Version in which the feature will be removed.
    replacement : callable | str
        Either a verbal description, or a pointer to the direct replacement
        function which takes the same arguments. In the latter case, the
        replacement is automatically called and the deprecated function is not
        used anymore.
    """
    def dec(func):
        msg = f"{func.__name__} is deprecated and will be removed in version {version}"
        if isinstance(replacement, str):
            msg += '; ' + replacement
            call_func = func
        elif replacement is not None:
            msg += f"; use {replacement.__name__} instead"
            call_func = replacement
        else:
            raise TypeError(f"{replacement=}")
        func.__doc__ = msg

        @functools.wraps(func)
        def new(*args, **kwargs):
            warn(msg, DeprecationWarning)
            return call_func(*args, **kwargs)

        return new
    return dec


def _deprecated_alias(alias, for_, version):
    """Create a deprecated alias

    Parameters
    ----------
    alias : str
        Name of the alias.
    for_ : callable
        New function (replacement).
    version : str
        Version in which the feature will be removed.
    """
    @functools.wraps(for_)
    def new(*args, **kwargs):
        warn(f"{alias} has been renamed to {for_.__name__}. {alias} will be removed in version {version}", DeprecationWarning)
        return for_(*args, **kwargs)
    return new


def deprecated_attribute(version, class_name, replacement):
    if not isinstance(replacement, str):
        raise TypeError("replacement=%r" % (replacement,))

    class Dec:

        def __init__(self, meth):
            self._meth = meth
            self.__name__ = meth.__name__
            self._msg = (
                'The %s.%s attribute is deprecated and will be removed in '
                'version %s, use %s.%s instead.' %
                (class_name, meth.__name__, version, class_name, replacement))

        def __get__(self, obj, klass=None):
            if obj is None:
                return None
            warn(self._msg, DeprecationWarning)
            return getattr(obj, replacement)

    return Dec


def deprecate_kwarg(
        old: str,
        new: str,
        since: str = None,
        remove_in: str = None,
):
    """
    Decorator factory to deprecate a renamed keyword argument.

    Parameters
    ----------
    old
        The deprecated keyword name that used to be accepted.
    new
        The new keyword name the function now uses.
    since
        Version/date when the deprecation started (for the warning message).
    remove_in
        Version/date when support will be removed (for the warning message).

    Notes
    -----
    - If both `old` and `new` are provided in a call, a TypeError is raised.
    - If `old` is used, it is moved to `new` and a DeprecationWarning is emitted.
    - Positional arguments are unaffected (this handles keyword renames).
    """
    def _decorator(func):
        sig = inspect.signature(func)
        if new not in sig.parameters:
            raise ValueError(f"{new=}: {func.__name__!r} has no parameter named {new!r}")

        @functools.wraps(func)
        def _wrapper(*args, **kwargs):
            if old in kwargs:
                if new in kwargs:
                    raise TypeError(f"{func.__name__} received both {old!r} (deprecated) and {new!r}. Use only {new!r}.")
                # Build a helpful warning message.
                parts = [f"Argument {old!r} is deprecated; use {new!r} instead."]
                if since:
                    parts.append(f"Deprecated since eelbrain {since}.")
                if remove_in:
                    parts.append(f"Support will be removed in eelbrain {remove_in}.")
                warn(" ".join(parts), DeprecationWarning, stacklevel=2)

                # Move value from old -> new
                kwargs[new] = kwargs.pop(old)

            # Let Python handle defaults/type errors as usual
            bound = sig.bind_partial(*args, **kwargs)
            return func(*bound.args, **bound.kwargs)

        return _wrapper
    return _decorator


def deprecate_ds_arg(func):
    """Backwards compatibility and deprecation for the `ds` parameter (renamed to `data`)"""
    @functools.wraps(func)
    def new(*args, ds=None, **kwargs):
        if ds is not None:
            warn("The `ds` argument has been renamed to `data`. `ds` will stop working in Eelbrain 0.41.", DeprecationWarning)
            kwargs['data'] = ds
        return func(*args, **kwargs)
    return new


def log_level(arg):
    """Convert string to logging module constant"""
    if isinstance(arg, int):
        return arg
    elif isinstance(arg, str):
        try:
            return LOG_LEVELS[arg.upper()]
        except KeyError:
            raise ValueError("Invalid log level: %s. mus be one of %s" %
                             (arg, ', '.join(LOG_LEVELS)))
    else:
        raise TypeError("Invalid log level: %s. need int or str." % repr(arg))


def set_log_level(level, logger_name='eelbrain'):
    """Set the minimum level of messages to be logged

    Parameters
    ----------
    level : str | int
        Level as string (debug, info, warning, error, critical) or
        corresponding constant from the logging module.
    logger_name : str
        Name of the logger for which to set the logging level. The default is
        the Eelbrain logger.
    """
    logging.getLogger(logger_name).setLevel(log_level(level))


class ScreenHandler(logging.StreamHandler):
    "Log handler compatible with TQDM"

    def __init__(self, formatter=None):
        logging.StreamHandler.__init__(self)
        if formatter is None:
            formatter = logging.Formatter("%(levelname)-8s:  %(message)s")
        self.setFormatter(formatter)

    def emit(self, record):
        tqdm.write(self.format(record))


def intervals(seq, first=None):
    """Iterate over each successive pair in a sequence.

    Examples
    --------
    >>> for i in intervals([1, 2, 3, 45]):
    ...     print(i)
    ...
    (1, 2)
    (2, 3)
    (3, 45)
    """
    iterator = iter(seq)
    if first is None:
        try:
            last = next(iterator)
        except StopIteration:
            return
    else:
        last = first

    for item in iterator:
        yield last, item
        last = item


@dataclass
class PickleableDataClass:

    def __getstate__(self) -> dict:
        return {f.name: getattr(self, f.name) for f in fields(self)}

    def __setstate__(self, state: dict):
        self.__init__(**state)


def _natural_keys(text):
    "Sorting key for natural sorting"
    # after http://stackoverflow.com/a/5967539/166700
    return [int(c) if c.isdigit() else c for c in re.split(r'(\d+)', text)]


def natsorted(seq):
    return sorted(seq, key=_natural_keys)


def n_decimals(number):
    "Number of meaningful decimals in ``number``, at least 1"
    s = str(number)
    if '.' in s:
        return s[::-1].index('.')
    else:
        return 1


class keydefaultdict(defaultdict):  # noqa
    "http://stackoverflow.com/a/2912455/166700"
    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        else:
            ret = self[key] = self.default_factory(key)
            return ret



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
import warnings

import numpy as np
from numpy.testing import assert_allclose, assert_equal
import pytest
import scipy.stats

from eelbrain import datasets, Model, Var
from eelbrain._stats import stats
from eelbrain._stats.permutation import permute_order, rand_rotation_matrices
from eelbrain._utils.r_bridge import r


def test_corr():
    "Test stats.corr"
    ds = datasets.get_uts()
    y = ds.eval("uts.x[:,:3]")
    x = ds.eval('Y.x')
    n_cases = len(y)
    df = n_cases - 2

    corr = stats.corr(y, x)
    p = stats.rtest_p(corr, df)
    for i in range(len(corr)):
        r_sp, p_sp = scipy.stats.pearsonr(y[:, i], x)
        assert corr[i] == pytest.approx(r_sp)
        assert p[i] == pytest.approx(p_sp)

    # NaN
    with warnings.catch_warnings():  # divide by 0
        warnings.simplefilter("ignore")
        assert stats.corr(np.arange(10), np.zeros(10)) == 0

    # perm
    y_perm = np.empty_like(y)
    for perm in permute_order(n_cases, 2):
        y_perm[perm] = y
        stats.corr(y, x, corr, perm)
        for i in range(len(corr)):
            r_sp, _ = scipy.stats.pearsonr(y_perm[:, i], x)
            assert corr[i] == pytest.approx(r_sp)


def test_lm():
    "Test linear model function against scipy lstsq"
    ds = datasets.get_uts(True)
    uts = ds['uts']
    utsnd = ds['utsnd']
    x = ds.eval("A*B")
    p = x._parametrize()
    n = ds.n_cases

    # 1d betas
    betas = stats.lm_betas(uts.x, p)
    sp_betas = scipy.linalg.lstsq(p.x, uts.x.reshape((n, -1)))[0]
    assert_allclose(betas, sp_betas)

    # 2d betas
    betas = stats.lm_betas(utsnd.x, p)
    sp_betas = scipy.linalg.lstsq(p.x, utsnd.x.reshape((n, -1)))[0]
    sp_betas = sp_betas.reshape((x.df,) + utsnd.shape[1:])
    assert_allclose(betas, sp_betas)


def test_lm_r():
    "Test linear model agains R"
    ds = datasets.get_uv()
    ds['A2'] = Var(ds['A'] == 'a2').astype(float)
    ds['B2'] = Var(ds['B'] == 'b2').astype(float)
    ds.to_r('ds')
    y = ds['fltvar'].x[:, None]

    p = Model([ds['A2']])._parametrize('dummy')
    b, se, t = stats.lm_t(y, p)
    r_res = r("summary(lm(fltvar ~ A2, ds, type=2))")
    assert_allclose(b[:, 0], r_res[3][0:2])
    assert_allclose(se[:, 0], r_res[3][2:4])
    assert_allclose(t[:, 0], r_res[3][4:6])

    p = ds.eval("A2 + B2")._parametrize('dummy')
    b, se, t = stats.lm_t(y, p)
    r_res = r("summary(lm(fltvar ~ A2 + B2, ds, type=2))")
    assert_allclose(b[:, 0], r_res[3][0:3])
    assert_allclose(se[:, 0], r_res[3][3:6])
    assert_allclose(t[:, 0], r_res[3][6:9])

    p = ds.eval("A2 + B2 + intvar")._parametrize('dummy')
    b, se, t = stats.lm_t(y, p)
    r_res = r("summary(lm(fltvar ~ A2 + B2 + intvar, ds, type=2))")
    assert_allclose(b[:, 0], r_res[3][0:4])
    assert_allclose(se[:, 0], r_res[3][4:8])
    assert_allclose(t[:, 0], r_res[3][8:12])


def test_dispersion():
    "Test variability functions"
    ds = datasets.get_loftus_masson_1994()
    y = ds['n_recalled'].x.astype(np.float64)
    x = ds['exposure'].as_factor()
    match = ds['subject']

    sem = scipy.stats.sem(y, 0, 1)
    ci = sem * scipy.stats.t.isf(0.05 / 2., len(y) - 1)

    # invalid spec
    with pytest.raises(ValueError):
        stats.dispersion(y, 0, 0, '1mile')
    with pytest.raises(ValueError):
        stats.dispersion(y, 0, 0, 'ci7ci')

    # standard error
    assert stats.dispersion(y, None, None, 'sem') == sem
    assert stats.dispersion(y, None, None, '2sem') == 2 * sem
    # within subject standard-error
    target = scipy.stats.sem(stats.residuals(y[:, None], match), 0, len(match.cells))[0]
    assert stats.dispersion(y, None, match, 'sem') == pytest.approx(target)
    # one data point per match cell
    n = match.df + 1
    with pytest.raises(ValueError):
        stats.dispersion(y[:n], None, match[:n], 'sem')

    target = np.array([scipy.stats.sem(y[x == cell], 0, 1) for cell in x.cells])
    es = stats.dispersion(y, x, None, 'sem')
    assert_allclose(es, target)

    stats.dispersion(y, x, None, 'sem', pool=True)

    # confidence intervals
    assert stats.dispersion(y, None, None, '95%ci') == pytest.approx(ci)
    assert stats.dispersion(y, x, None, '95%ci', pool=True) == pytest.approx(3.86, abs=1e-2)  # L&M: 3.85
    assert stats.dispersion(y, x, match, '95%ci') == pytest.approx(0.52, abs=1e-2)

    assert_equal(
        stats.dispersion(y, x, None, '95%ci')[::-1],
        stats.dispersion(y, x, None, '95%ci', x.cells[::-1])
    )


def test_t_1samp():
    "Test 1-sample t-test"
    ds = datasets.get_uts(True)

    y = ds['uts'].x
    t = scipy.stats.ttest_1samp(y, 0, 0)[0]
    assert_allclose(stats.t_1samp(y), t, 10)

    y = ds['utsnd'].x
    t = scipy.stats.ttest_1samp(y, 0, 0)[0]
    assert_allclose(stats.t_1samp(y), t, 10)


def test_t_ind():
    "Test independent samples t-test"
    ds = datasets.get_uts(True)
    y = ds.eval("utsnd.x")
    n_cases = len(y)
    n = n_cases // 2
    groups = (np.arange(n_cases) < n)
    groups.dtype = np.int8

    t = stats.t_ind(y, groups)
    p = stats.ttest_p(t, n_cases - 2)
    t_sp, p_sp = scipy.stats.ttest_ind(y[:n], y[n:])
    assert_allclose(t, t_sp)
    assert_allclose(p, p_sp)
    assert_allclose(stats.ttest_t(p, n_cases - 2), np.abs(t))

    # permutation
    y_perm = np.empty_like(y)
    for perm in permute_order(n_cases, 2):
        stats.t_ind(y, groups, out=t, perm=perm)
        y_perm[perm] = y
        t_sp, _ = scipy.stats.ttest_ind(y_perm[:n], y_perm[n:])
        assert_allclose(t, t_sp)


def test_vector():
    rotation = rand_rotation_matrices(4, 0)
    zeros = np.zeros((4, 3))
    assert stats.t2_1samp(zeros) == 0
    assert stats.t2_1samp(zeros, rotation) == 0



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Pre-processing operations based on NDVars"""
from __future__ import annotations
import warnings
from copy import deepcopy
import fnmatch
from itertools import chain
import logging
from os import makedirs, remove
from os.path import basename, dirname, exists, getmtime, join, splitext
from pathlib import Path
from typing import Any, Collection, Dict, List, Literal, Sequence, Tuple, Union

import mne
from scipy import signal
from mne_bids import BIDSPath, read_raw_bids, mark_channels
import pandas as pd

from .. import load
from .._data_obj import NDVar, Sensor
from .._exceptions import DefinitionError
from .._io.fiff import KIT_NEIGHBORS
from .._io.txt import read_adjacency
from .._ndvar import filter_data
from .._text import enumeration
from .._utils import ask, deprecate_kwarg, user_activity
from ..mne_fixes import CaptureLog
from ..mne_fixes._version import MNE_VERSION, V0_19, V0_24
from .definitions import log_dict_change, tuple_arg, typed_arg
from .exceptions import FileMissingError

AddBadsArg = Union[bool, Sequence[str]]
PreloadArg = Union[bool, Literal[-1]]


class RawPipe:
    name: str = None  # set on linking
    log: logging.Logger = None

    def _can_link(self, pipes: Dict[str, RawPipe]) -> bool:
        raise NotImplementedError

    def _link(
            self,
            name: str,
            pipes: Dict[str, RawPipe],
            cache_dir: str,
            deriv_dir: str,
            log: logging.Logger
    ) -> RawPipe:
        raise NotImplementedError

    def _link_base(
            self,
            name: str,
            log: logging.Logger,
    ) -> RawPipe:
        out = deepcopy(self)
        out.name = name
        out.log = log
        return out

    def _as_dict(self, args: Sequence[str] = ()) -> dict:
        out = {arg: getattr(self, arg) for arg in chain(args, ('name',))}
        out['type'] = self.__class__.__name__
        return out

    @staticmethod
    def _normalize_dict(state: dict) -> None:
        pass

    def cache(self, path: BIDSPath) -> mne.io.BaseRaw:
        "Make sure the file exists and is up to date"
        raise NotImplementedError

    def get_adjacency(self, data: str) -> Union[str, List[Tuple[str, str]], Path]:
        raise NotImplementedError

    def get_sysname(
            self,
            info: mne.Info,
            subject: str,
            data: str
    ) -> str | None:
        raise NotImplementedError

    def get_path(
            self,
            path: BIDSPath,
            file_type: Literal['raw', 'bads', 'cache', 'ica'] = 'raw',
            on_error: Literal['raise', 'warn', 'ignore'] = 'raise',
    ) -> str | None:
        if not path.find_matching_sidecar(on_error=on_error):
            return None
        if file_type == 'raw':
            return str(path.fpath)
        elif file_type == 'bads':
            return str(path.find_matching_sidecar('channels', '.tsv', on_error=on_error))
        elif file_type == 'cache':
            return join(
                self.cache_dir,
                'raw',
                get_subject_session(path.basename),
                f'{splitext(path.basename)[0]}_raw-{self.name}.fif'
            )
        else:
            return join(
                self.deriv_dir,
                'ica',
                f'{remove_task(splitext(path.basename)[0])}_raw-{self.name}_ica.fif'
            )

    def load(
            self,
            path: BIDSPath,
            add_bads: AddBadsArg = True,
            preload: PreloadArg = False,  # -1: info only, data will not be needed
            raw: mne.io.BaseRaw = None,
    ) -> mne.io.BaseRaw:
        # raw
        if raw is None:
            if preload == -1:
                preload = False
            raw = self._load(path, preload)
        # bad channels
        if isinstance(add_bads, Sequence):
            raw.info['bads'] = list(add_bads)
        elif add_bads is False:
            raw.info['bads'] = []
        elif add_bads is not True:
            raise TypeError(f"{add_bads=}")
        return raw

    def _load(
            self,
            path: BIDSPath,
            preload: PreloadArg,
    ) -> mne.io.BaseRaw:
        return read_raw_bids(
            path,
            extra_params={'preload': preload},
            verbose='ERROR',
        )

    def load_bad_channels(
            self,
            path: BIDSPath,
            existing: Collection[str] = None,
    ) -> list[str]:
        raise NotImplementedError

    def make_bad_channels(
            self,
            path: BIDSPath,
            bad_chs,
            redo: bool,
    ) -> None:
        raise NotImplementedError

    def make_bad_channels_auto(
            self,
            path: BIDSPath,
            flat: float,
            redo: bool,
    ) -> None:
        raise NotImplementedError

    def mtime(
            self,
            path: BIDSPath,
            bad_chs: bool = True,
    ) -> float | None:
        "Modification time of anything influencing the output of load"
        raise NotImplementedError


class RawSource(RawPipe):
    """Raw data source

    Parameters
    ----------
    filename
        Pattern for filenames. The pattern should contain the fields
        ``{subject}`` and ``{recording}`` (which internally is expanded to
        ``session`` and, if applicable, ``visit``;
        default ``'{subject}_{recording}-raw.fif'``).
    reader
        Function for reading data (default is :func:`mne.io.read_raw_fif`).
    sysname
        Used to determine sensor positions (not needed for KIT files, or when a
        montage is specified).
    rename_channels
        Rename channels based on a ``{from: to}`` dictionary. This happens
        *after* calling the ``reader``, and *before* applying the ``montage``.
        Useful to convert system-specific channel names to those of a standard montages.
    montage
        Name of a montage that is applied to raw data to set sensor positions
        (see :meth:`mne.io.Raw.set_montage`).
    adjacency
        Ajacency between sensors. Can be specified as:

        - ``'auto'`` to use :func:`mne.channels.find_ch_adjacency`
        - Pre-defined adjacency (one of :func:`mne.channels.get_builtin_ch_adjacencies`)
        - Path to load adjacency from a file
        - ``"none"`` for no connections
        - ``"grid"`` for grid connections
        - list of connections (e.g., ``[('OZ', 'O1'), ('OZ', 'O2'), ...]``)
        - :class:`numpy.ndarray` of int, shape (n_edges, 2), to specify
          connections in terms of indices. Each row should specify one
          connection [i, j] with i < j. If the array's dtype is uint32,
          property checks are disabled to improve efficiency.

        If unspecified, it is inferred from ``sysname`` if possible.
    ...
        Additional parameters for the ``reader`` function.

    See Also
    --------
    MneExperiment.raw

    Examples
    --------
    If you would load the EEG data like this::

        form pathlib import Path

        subject = '0001'
        raw = mne.io.read_raw_brainvision(
            f'/data/eeg/{subject}/raw/task_{subject}.vhdr',
            eog=['EOG horizontal left', 'EOG horizontal right', 'EOG vertical below', 'EOG vertical above'],
        )
        raw.rename_channels({'FZ': 'Fz', 'OZ': 'Oz', 'CZ': 'Cz', 'Ref right mastoid': 'A1'})
        raw.set_montage('standard_1020')

    Then the equivalent :class:`RawSource` is::

        raw = {
            'raw': RawSource(
                'raw/{recording}_{subject}.vhdr',
                reader=mne.io.read_raw_brainvision,
                rename_channels={'FZ': 'Fz', 'OZ': 'Oz', 'CZ': 'Cz', 'Ref right mastoid': 'A1'},
                eog=['EOG horizontal left', 'EOG horizontal right', 'EOG vertical below', 'EOG vertical above'],
                montage='standard_1020'),
            ...
        }

    """

    @deprecate_kwarg('connectivity', 'adjacency', '0.41', '0.42')
    def __init__(
            self,
            sysname: str = None,
            rename_channels: dict = None,
            montage: str = None,
            adjacency: Union[str, List[Tuple[str, str]], Path] = None,
            **kwargs,
    ):
        RawPipe.__init__(self)
        if isinstance(adjacency, str):
            if adjacency not in ('auto', 'grid', 'none') and adjacency not in mne.channels.get_builtin_ch_adjacencies():
                adjacency = Path(adjacency)
        if isinstance(adjacency, Path):
            adjacency = read_adjacency(adjacency)
        # TODO: Using MNE-BIDS to read raw limits the file formats, add more readers later
        # self.filename = typed_arg(filename, str)
        # self.reader = reader
        self.sysname = sysname
        self.rename_channels = typed_arg(rename_channels, dict)
        self.montage = montage
        self.adjacency = adjacency
        self._kwargs = kwargs
        # TODO: Should we keep this?
        # if MNE_VERSION < V0_19 and reader is mne.io.read_raw_cnt:
        #     self._read_raw_kwargs = {'montage': None, **kwargs}
        # else:
        #     self._read_raw_kwargs = kwargs

    def _can_link(self, pipes: Dict[str, RawPipe]) -> bool:
        return True

    def _link(
            self,
            name: str,
            pipes: Dict[str, RawPipe],
            cache_dir: str,
            deriv_dir: str,
            log: logging.Logger,
    ) -> RawPipe:
        return RawPipe._link_base(self, name, log)

    def _as_dict(self, args: Sequence[str] = ()) -> dict:
        out = RawPipe._as_dict(self, args)
        out.update(self._kwargs)
        if self.rename_channels:
            out['rename_channels'] = self.rename_channels
        if self.montage:
            if isinstance(self.montage, mne.channels.DigMontage):
                out['montage'] = Sensor.from_montage(self.montage)
            else:
                out['montage'] = self.montage
        if self.adjacency is not None:
            out['connectivity'] = self.adjacency
        return out

    def _load(self, path: BIDSPath, preload):
        raw: mne.io.BaseRaw = read_raw_bids(
            path,
            extra_params={'preload': preload},
            verbose='ERROR',
        )
        if self.rename_channels:
            if rename := {k: v for k, v in self.rename_channels.items() if k in raw.ch_names}:
                raw.rename_channels(rename)
        if self.montage:
            raw.set_montage(self.montage)
        # Empty room may have missing raw.info['dig'], find alternative ways to do this when encountered
        # if not raw.info['dig'] and self._dig_sessions is not None and self._dig_sessions[subject]:
        #     dig_recording = self._dig_sessions[subject][recording]
        #     if dig_recording != recording:
        #         dig_raw = self._load(subject, dig_recording, False)
        #         raw.set_montage(mne.channels.DigMontage(dig=dig_raw.info['dig']))
        return raw

    def get_adjacency(self, data: str) -> Union[str, List[Tuple[str, str]], Path]:
        if data == 'eog':
            return None
        else:
            return self.adjacency

    def get_sysname(
            self,
            info: mne.Info,
            subject: str,
            data: str,
    ) -> str | None:
        if data == 'eog':
            return None
        elif isinstance(self.sysname, str):
            return self.sysname
        elif isinstance(self.sysname, dict):
            for k, v in self.sysname.items():
                if fnmatch.fnmatch(subject, k):
                    return v
        kit_system_id = info.get('kit_system_id')
        return KIT_NEIGHBORS.get(kit_system_id)

    def load_bad_channels(
            self,
            path: BIDSPath,
            existing: Collection[str] = None,
    ) -> list[str]:
        bads_path = self.get_path(path, 'bads')
        channels_df = pd.read_csv(bads_path, sep='\t')
        if (channels_df is None) or ('status' not in channels_df.columns.tolist()):
            self.log.info("Generating bad_channels for %s", path.fpath)
            self.make_bad_channels_auto(path)
        bad_chs = channels_df.query('status == "bad"')['name'].tolist()
        if existing is not None:
            bad_chs = [ch for ch in bad_chs if ch in existing]
        return bad_chs

    def make_bad_channels(
            self,
            path: BIDSPath,
            bad_chs,
            redo: bool,
    ) -> None:
        bads_path = self.get_path(path, 'bads')
        channels_df = pd.read_csv(bads_path, sep='\t')
        if (channels_df is None) or ('status' not in channels_df.columns.tolist()):
            old_bads = None
        else:
            old_bads = channels_df.query('status == "bad"')['name'].tolist()
        # find new bad channels
        if isinstance(bad_chs, (str, int)):
            bad_chs = (bad_chs,)
        raw = self.load(path, add_bads=False)
        sensor = load.mne.sensor_dim(raw.info, adjacency=self.adjacency)
        new_bads = sensor._normalize_sensor_names(bad_chs)
        # update with old bad channels
        if old_bads is not None and not redo:
            new_bads = sorted(set(old_bads).union(new_bads))
        # print change
        print(f"{old_bads} -> {new_bads}")
        if new_bads == old_bads:
            return
        # write new bad channels
        mark_channels(path, ch_names=new_bads, status='bad')

    def make_bad_channels_auto(
            self,
            path: BIDSPath,
            flat: float = None,
            redo: bool = False,
    ) -> None:
        raw = self.load(path, add_bads=False)
        bad_chs: list[str] = raw.info['bads']
        if flat is None:
            flat = 1e-14  # May need a setting to exclude the EEG reference?
        if flat:
            sysname = self.get_sysname(raw.info, path.entities['subject'], None)
            raw = load.mne.raw_ndvar(raw, sysname=sysname, adjacency=self.adjacency)
            bad_chs.extend(raw.sensor.names[raw.std('time') < flat])
        self.make_bad_channels(path, bad_chs, redo)

    def mtime(
            self,
            path: BIDSPath,
            bad_chs: bool = True,
    ) -> float | None:
        raw_path = self.get_path(path)
        bads_path = self.get_path(path, 'bads')
        if bad_chs:
            return max(getmtime(raw_path), getmtime(bads_path))
        else:
            return getmtime(raw_path)


class CachedRawPipe(RawPipe):
    _bad_chs_affect_cache: bool = False
    source: RawPipe = None  # set on linking

    def __init__(self, source, cache=True):
        RawPipe.__init__(self)
        self._source_name = source
        self._cache = cache

    def _can_link(self, pipes: Dict[str, RawPipe]) -> bool:
        return self._source_name in pipes

    def _link(
            self,
            name: str,
            pipes: Dict[str, RawPipe],
            cache_dir: str,
            deriv_dir: str,
            log: logging.Logger,
    ) -> RawPipe:
        if self._source_name not in pipes:
            raise DefinitionError(f"{self.__class__.__name__} {name!r} source {self._source_name!r} does not exist")
        out = RawPipe._link_base(self, name, log)
        out.source = pipes[self._source_name]
        out.cache_dir = cache_dir
        out.deriv_dir = deriv_dir
        return out

    def _as_dict(self, args: Sequence[str] = ()) -> dict:
        out = RawPipe._as_dict(self, args)
        out['source'] = self._source_name
        return out

    def cache(self, path: BIDSPath) -> mne.io.BaseRaw:
        "Make sure the cache is up to date"
        cache_path = self.get_path(path, 'cache')
        if exists(cache_path):
            mtime = self.mtime(path, self._bad_chs_affect_cache)
            if mtime and getmtime(cache_path) >= mtime:
                return
        from .. import __version__
        # make sure the target directory exists
        makedirs(dirname(cache_path), exist_ok=True)
        # generate new raw
        with CaptureLog(cache_path[:-3] + 'log') as logger:
            logger.info(f"eelbrain {__version__}")
            logger.info(f"mne {mne.__version__}")
            logger.info(repr(self._as_dict()))
            raw = self._make(path, True)
        # save
        try:
            raw.save(cache_path, overwrite=True, verbose='critical')
        except BaseException:
            # clean up potentially corrupted file
            if exists(cache_path):
                remove(cache_path)
            raise
        return raw

    def get_adjacency(self, data: str) -> Union[str, List[Tuple[str, str]], Path]:
        return self.source.get_adjacency(data)

    def get_sysname(
            self,
            info: mne.Info,
            subject: str,
            data: str,
    ) -> str | None:
        return self.source.get_sysname(info, subject, data)

    def load(
            self,
            path: BIDSPath,
            add_bads: AddBadsArg = True,
            preload: PreloadArg = False,  # -1: info only, data will not be needed
            raw: mne.io.BaseRaw = None,
    ) -> mne.io.BaseRaw:
        if raw is not None:
            pass
        elif self._cache:
            raw = self.cache(path)
        elif preload == -1:
            raw = self._make_info(path)
        else:
            raw = self._make(path, preload)
        return RawPipe.load(self, path, add_bads, preload, raw)

    def load_bad_channels(
            self,
            path: BIDSPath,
            existing: Collection[str] = None,
    ) -> list[str]:
        return self.source.load_bad_channels(path, existing)

    def _make(
            self,
            path: BIDSPath,
            preload: bool,
    ) -> mne.io.BaseRaw:
        raise NotImplementedError

    def _make_info(self, path: BIDSPath) -> mne.io.BaseRaw:
        return self.source.load(path, preload=-1)

    def make_bad_channels(
            self,
            path: BIDSPath,
            bad_chs,
            redo: bool,
    ) -> None:
        self.source.make_bad_channels(path, bad_chs, redo)

    def make_bad_channels_auto(self, *args, **kwargs) -> None:
        self.source.make_bad_channels_auto(*args, **kwargs)

    def mtime(
            self,
            path: BIDSPath,
            bad_chs: bool = True,
    ) -> float | None:
        return self.source.mtime(path, bad_chs or self._bad_chs_affect_cache)


class RawFilter(CachedRawPipe):
    """Filter raw pipe

    Parameters
    ----------
    source
        Name of the raw pipe to use for input data.
    l_freq
        Low cut-off frequency in Hz.
    h_freq
        High cut-off frequency in Hz.
    cache
        Cache the resulting raw files (default ``True``).
    n_jobs
        Parameter for :meth:`mne.io.Raw.filter`; Values other than 1 are slower
        in most cases due to added overhead except for very large files.
    ...
        :meth:`mne.io.Raw.filter` parameters.

    See Also
    --------
    MneExperiment.raw
    """
    def __init__(
            self,
            source: str,
            l_freq: float = None,
            h_freq: float = None,
            cache: bool = True,
            n_jobs: Union[str, int, None] = 1,
            **kwargs,
    ):
        CachedRawPipe.__init__(self, source, cache)
        self.args = (l_freq, h_freq)
        self.kwargs = kwargs
        self.n_jobs = n_jobs
        # mne backwards compatibility (fir_design default change 0.15 -> 0.16)
        if 'use_kwargs' in kwargs:
            self._use_kwargs = kwargs.pop('use_kwargs')
        else:
            self._use_kwargs = kwargs

    def _as_dict(self, args: Sequence[str] = ()) -> dict:
        return CachedRawPipe._as_dict(self, [*args, 'args', 'kwargs'])

    def filter_ndvar(self, ndvar, **kwargs):
        return filter_data(ndvar, *self.args, **self._use_kwargs, **kwargs)

    def _make(
            self,
            path: BIDSPath,
            preload: bool,
    ) -> mne.io.BaseRaw:
        raw = self.source.load(path, preload=True)
        self.log.info("Raw %s: filtering for %s...", self.name, path.fpath)
        raw.filter(*self.args, **self._use_kwargs, n_jobs=self.n_jobs)
        return raw

    def _make_info(self, path: BIDSPath) -> mne.io.BaseRaw:
        raw = super()._make_info(path)
        l_freq, h_freq = self.args
        if l_freq and l_freq > (raw.info['highpass'] or 0):
            with raw.info._unlock():
                raw.info['highpass'] = float(l_freq)
        if h_freq and h_freq < (raw.info['lowpass'] or raw.info['sfreq']):
            with raw.info._unlock():
                raw.info['lowpass'] = float(h_freq)
        return raw


class RawFilterElliptic(CachedRawPipe):

    def __init__(self, source, low_stop, low_pass, high_pass, high_stop, gpass, gstop):
        CachedRawPipe.__init__(self, source)
        self.args = (low_stop, low_pass, high_pass, high_stop, gpass, gstop)

    def _as_dict(self, args: Sequence[str] = ()) -> dict:
        return CachedRawPipe._as_dict(self, [*args, 'args'])

    def _sos(self, sfreq):
        nyq = sfreq / 2.
        low_stop, low_pass, high_pass, high_stop, gpass, gstop = self.args
        if high_stop is None:
            assert low_stop is not None
            assert high_pass is None
        else:
            high_stop /= nyq
            high_pass /= nyq

        if low_stop is None:
            assert low_pass is None
        else:
            low_pass /= nyq
            low_stop /= nyq

        if low_stop is None:
            btype = 'lowpass'
            wp, ws = high_pass, high_stop
        elif high_stop is None:
            btype = 'highpass'
            wp, ws = low_pass, low_stop
        else:
            btype = 'bandpass'
            wp, ws = (low_pass, high_pass), (low_stop, high_stop)
        order, wn = signal.ellipord(wp, ws, gpass, gstop)
        return signal.ellip(order, gpass, gstop, wn, btype, output='sos')

    def filter_ndvar(self, ndvar):
        axis = ndvar.get_axis('time')
        sos = self._sos(1. / ndvar.time.tstep)
        x = signal.sosfilt(sos, ndvar.x, axis)
        return NDVar(x, ndvar.dims, ndvar.info.copy(), ndvar.name)

    def _make(
            self,
            path: BIDSPath,
            preload: bool,
    ) -> mne.io.BaseRaw:
        raw = self.source.load(path, preload=True)
        self.log.info("Raw %s: filtering for %s...", self.name, path.fpath)
        # filter data
        picks = mne.pick_types(raw.info, meg=True, eeg=True, ref_meg=True)
        sos = self._sos(raw.info['sfreq'])
        for i in picks:
            raw._data[i] = signal.sosfilt(sos, raw._data[i])
        # update info
        low, high = self.args[1], self.args[2]
        if high and raw.info['lowpass'] > high:
            raw.info['lowpass'] = float(high)
        if low and raw.info['highpass'] < low:
            raw.info['highpass'] = float(low)
        return raw


class RawICA(CachedRawPipe):
    """ICA raw pipe

    Parameters
    ----------
    source
        Name of the raw pipe to use for input data.
    task
        Session(s) to use for estimating ICA components.
    method
        Method for ICA decomposition (default: ``'extended-infomax'``; see
        :class:`mne.preprocessing.ICA`).
    random_state
        Set the random state for ICA decomposition to make results reproducible
        (default 0, see :class:`mne.preprocessing.ICA`).
    fit_kwargs
        A dictionary with keyword arguments that should be passed to
        :meth:`mne.preprocessing.ICA.fit`. This includes
        ``reject={'mag': 5e-12, 'grad': 5000e-13, 'eeg': 300e-6}`` unless
        a different value for ``reject`` is specified here.
    cache : bool
        Cache the resulting raw files (default ``False``).
    ...
        Additional parameters for :class:`mne.preprocessing.ICA`.

    See Also
    --------
    MneExperiment.raw
    RawApplyICA

    Notes
    -----
    This preprocessing step estimates one set of ICA components per subject,
    using the data specified in the ``task`` parameter. The selected
    components are then removed from all data sessions during this preprocessing
    step, regardless of whether they were used to estimate the components or
    not.

    Use :meth:`MneExperiment.make_ica_selection` for each subject to
    select ICA components that should be removed. The arguments to that function
    determine what data is used to visualize the component time courses.
    For example, to determine which components load strongly on empty room data,
    use ``e.make_ica_selection(task='emptyroom')`` (assuming an
    ``'emptyroom'`` task is present).

    This step merges bad channels from all sessions.

    Examples
    --------
    Some ICA examples::

        class Experiment(MneExperiment):

            raw = {
                '1-40': RawFilter('raw', 1, 40),
                # Extended infomax with PCA preprocessing
                'ica': RawICA('1-40', 'extended-infomax', n_components=0.99),
                # Fast ICA
                'fastica': RawICA('1-40', 'task', 'fastica', n_components=0.9),
                # Change thresholds for data rejection using fit_kwargs
                'ica-rej': RawICA('1-40', 'task', 'fastica', fit_kwargs=dict(
                    reject={'mag': 5e-12, 'grad': 5000e-13, 'eeg': 500e-6},
                )),
            }

    """

    def __init__(
            self,
            source: str,
            task: Union[str, Sequence[str]],
            method: str = 'extended-infomax',
            random_state: int = 0,
            fit_kwargs: Dict[str, Any] = None,
            cache: bool = False,
            **kwargs,
    ):
        CachedRawPipe.__init__(self, source, cache)
        self.task = tuple_arg('task', task, allow_none=False)
        self.kwargs = {'method': method, 'random_state': random_state, **kwargs}
        self.fit_kwargs = dict(fit_kwargs) if fit_kwargs else {}

    def _as_dict(self, args: Sequence[str] = ()) -> dict:
        out = CachedRawPipe._as_dict(self, [*args, 'task', 'kwargs'])
        if self.fit_kwargs:
            out['fit_kwargs'] = self.fit_kwargs
        return out

    def load_bad_channels(
            self,
            path: BIDSPath,
            existing: Collection[str] = None,
    ) -> list[str]:
        bad_chs = set()
        for task in self.task:
            path.copy().update(task=task)
            bad_chs.update(self.source.load_bad_channels(path, existing))
        return sorted(bad_chs)

    def load_ica(self, path: BIDSPath) -> mne.preprocessing.ICA:
        ica_path = self.get_path(path, 'ica')
        if not exists(ica_path):
            raise FileMissingError(f"ICA file {basename(ica_path)} does not exist for raw={self.name!r}. Run e.make_ica_selection() to create it.")
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', 'Version 0.23 introduced max_iter', DeprecationWarning)
            return mne.preprocessing.read_ica(ica_path)

    @staticmethod
    def _check_ica_channels(
            ica: mne.preprocessing.ICA,
            raw: mne.io.BaseRaw,
            raise_on_mismatch: bool = False,
            raw_name: str = None,
            subject: str = None,
            return_missing: bool = False,  # if ICA is only missing channels, retrun those
    ) -> bool | tuple:
        picks = mne.pick_types(raw.info, meg=True, eeg=True, ref_meg=False)
        raw_ch_names = [raw.ch_names[i] for i in picks]
        names_match = ica.ch_names == raw_ch_names
        if return_missing:
            assert raise_on_mismatch
            if names_match:
                return ()
            ica_set = set(ica.ch_names)
            raw_set = set(raw_ch_names)
            if not ica_set - raw_set:
                if missing := raw_set - ica_set:
                    return tuple(missing)
        if raise_on_mismatch and not names_match:
            raise RuntimeError(f"The ICA channel names do not match the data channels for raw={raw_name!r}, {subject=}. Have the bad channels changed since the ICA was computed? Try to revert the data channels, or recompute the ICA using MneExperiment.make_ica().\nData: {', '.join(raw_ch_names)}\nICA:  {', '.join(ica.ch_names)}")
        return names_match

    def load_concatenated_source_raw(
            self,
            path: BIDSPath,
            tasks: list[str],
            runs: list[str],
    ) -> mne.io.BaseRaw:
        bad_channels = self.load_bad_channels(path)
        path_list = []
        for task in tasks:
            path_ = path.copy().update(task=task)
            if not runs:
                path_list.append(path_)
                continue
            for run in runs:
                path_list.append(path_.copy().update(run=run))
        raw = self.source.load(path_list[0], bad_channels)
        for path_ in path_list[1:]:
            raw_ = self.source.load(path_, bad_channels)
            raw.append(raw_)
        return raw

    def make_ica(
            self,
            path: BIDSPath,
            run: list[str],
    ) -> str:
        ica_path = self.get_path(path, 'ica')
        bad_channels = self.load_bad_channels(path)
        if exists(ica_path):
            raw = self.source.load(path.copy().update(task=self.task[0]), bad_channels, preload=-1)
            ica = mne.preprocessing.read_ica(ica_path)
            # equal channel names in different raw is guaranteed here
            if not self._check_ica_channels(ica, raw):
                self.log.info("Raw %s for subject=%r: ICA channels mismatch data channels, recomputing ICA...", self.name, path.entities['subject'])
            else:
                mtimes = [self.source.mtime(path.copy().update(task=task), self._bad_chs_affect_cache) for task in self.task]
                if all(mtimes) and getmtime(ica_path) > max(mtimes):
                    return ica_path
                # ICA file is newer than raw
                command = ask(f"The input for the ICA of {path.entities['subject']} seems to have changed since the ICA was generated.", {'delete': 'delete and recompute the ICA', 'ignore': 'Keep using the old ICA'}, help="This message indicates that the modification date of the raw input data or of the bad channels file is more recent than that of the ICA file. If the data actually changed, ICA components might not be valid anymore and should be recomputed. If the change is spurious (e.g., the raw file was modified in a way that does not affect the ICA) load and resave the ICA file to stop seeing this message.")
                if command == 'ignore':
                    return ica_path
                elif command == 'delete':
                    remove(ica_path)
                else:
                    raise RuntimeError(f"{command=}")

        raw = self.load_concatenated_source_raw(path, self.task, run)
        self.log.info("Raw %s: computing ICA decomposition for %s", self.name, path.entities['subject'])
        kwargs = self.kwargs.copy()
        kwargs.setdefault('max_iter', 256)
        if MNE_VERSION > V0_19 and kwargs['method'] == 'extended-infomax':
            kwargs['method'] = 'infomax'
            kwargs['fit_params'] = {'extended': True}

        ica = mne.preprocessing.ICA(**kwargs)
        # reject presets from meeg-preprocessing
        fit_kwargs = {'reject': {'mag': 5e-12, 'grad': 5000e-13, 'eeg': 300e-6}, **self.fit_kwargs}
        with user_activity:
            ica.fit(raw, **fit_kwargs)
        makedirs(dirname(ica_path), exist_ok=True)
        if MNE_VERSION >= V0_24:
            ica.save(ica_path, overwrite=True)
        else:
            ica.save(ica_path)
        return ica_path

    def _make(
            self,
            path: BIDSPath,
            preload: PreloadArg,
    ) -> mne.io.BaseRaw:
        raw = self.source.load(path, preload=True)
        return self._apply(raw, path, self.name)

    def _apply(
            self,
            raw: mne.io.BaseRaw,
            path: BIDSPath,
            raw_name: str,
    ) -> mne.io.BaseRaw:
        self.log.debug("Raw %s: applying ICA for %s...", raw_name, path.fpath)
        raw.info['bads'] = self.load_bad_channels(path, existing=raw.ch_names)
        ica = self.load_ica(path)
        missing = self._check_ica_channels(ica, raw, raise_on_mismatch=True, raw_name=raw_name, subject=path.entities['subject'], return_missing=True)
        if missing:
            raw.drop_channels(missing)
        ica.apply(raw)
        return raw

    def mtime(
            self,
            path: BIDSPath,
            bad_chs: bool = True,
    ) -> float | None:
        mtime = CachedRawPipe.mtime(self, path, bad_chs or self._bad_chs_affect_cache)
        if mtime:
            ica_path = self.get_path(path, 'ica')
            if exists(ica_path):
                return max(mtime, getmtime(ica_path))


class RawApplyICA(CachedRawPipe):
    """Apply ICA estimated in a :class:`RawICA` pipe

    Parameters
    ----------
    source
        Name of the raw pipe to use for input data.
    ica
        Name of the :class:`RawICA` pipe from which to load the ICA components.
    cache
        Cache the resulting raw files (default ``False``).

    See Also
    --------
    MneExperiment.raw

    Notes
    -----
    This pipe inherits bad channels from the ICA.

    Examples
    --------
    Estimate ICA components with 1-40 Hz band-pass filter and apply the ICA
    to data that is high pass filtered at 0.1 Hz::

        class Experiment(MneExperiment):

            raw = {
                '1-40': RawFilter('raw', 1, 40),
                'ica': RawICA('1-40', 'task', 'extended-infomax', n_components=0.99),
                '0.1-40': RawFilter('raw', 0.1, 40),
                '0.1-40-ica': RawApplyICA('0.1-40', 'ica'),
            }

    """
    ica_source = None  # set on linking

    def __init__(
            self,
            source: str,
            ica: str,
            cache: bool = False,
    ):
        CachedRawPipe.__init__(self, source, cache)
        self._ica_source = ica

    def _can_link(self, pipes: Dict[str, RawPipe]) -> bool:
        return CachedRawPipe._can_link(self, pipes) and self._ica_source in pipes

    def _link(
            self,
            name: str,
            pipes: Dict[str, RawPipe],
            cache_dir: str,
            deriv_dir: str,
            log: logging.Logger,
    ) -> RawPipe:
        out = CachedRawPipe._link(self, name, pipes, cache_dir, deriv_dir, log)
        out.ica_source = pipes[self._ica_source]
        return out

    def _as_dict(self, args: Sequence[str] = ()) -> dict:
        out = CachedRawPipe._as_dict(self, args)
        out['ica_source'] = self._ica_source
        return out

    def load_bad_channels(
            self,
            path: BIDSPath,
            existing: Collection[str] = None,
    ) -> list[str]:
        return self.ica_source.load_bad_channels(path, existing)

    def _make(
            self,
            path: BIDSPath,
            preload: PreloadArg,
    ) -> mne.io.BaseRaw:
        raw = self.source.load(path, preload=True)
        return self.ica_source._apply(raw, path, self.name)

    def mtime(
            self,
            path: BIDSPath,
            bad_chs: bool = True,
    ) -> float | None:
        mtime = CachedRawPipe.mtime(self, path, bad_chs)
        if mtime:
            ica_mtime = self.ica_source.mtime(path, bad_chs)
            if ica_mtime:
                return max(mtime, ica_mtime)


class RawMaxwell(CachedRawPipe):
    """Maxwell filter raw pipe

    Parameters
    ----------
    source : str
        Name of the raw pipe to use for input data.
    bad_condition : str
        How to deal with ill-conditioned SSS matrices; by default, an error is
        raised, which might prevent the process to complete for some subjects.
        Set to ``'warning'`` to proceed anyways.
    cache : bool
        Cache the resulting raw files (default ``True``).
    ...
        :func:`mne.preprocessing.maxwell_filter` parameters.

    See Also
    --------
    MneExperiment.raw
    """

    _bad_chs_affect_cache = True

    def __init__(self, source, bad_condition='error', cache=True, **kwargs):
        CachedRawPipe.__init__(self, source, cache)
        self.kwargs = kwargs
        self.bad_condition = bad_condition

    def _as_dict(self, args: Sequence[str] = ()) -> dict:
        return CachedRawPipe._as_dict(self, [*args, 'kwargs'])

    def _make(
            self,
            path: BIDSPath,
            preload: PreloadArg,
    ) -> mne.io.BaseRaw:
        raw = self.source.load(path)
        self.log.info("Raw %s: computing Maxwell filter for %s", self.name, path.fpath)
        with user_activity:
            return mne.preprocessing.maxwell_filter(raw, bad_condition=self.bad_condition, **self.kwargs)


class RawOversampledTemporalProjection(CachedRawPipe):
    """Oversampled temporal projection: see :func:`mne.preprocessing.oversampled_temporal_projection`"""

    def __init__(
            self,
            source: str,
            duration: float = 10.0,
            cache: bool = True,
    ):
        CachedRawPipe.__init__(self, source, cache)
        self.duration = duration

    def _as_dict(self, args: Sequence[str] = ()) -> dict:
        return CachedRawPipe._as_dict(self, [*args, 'duration'])

    def _make(
            self,
            path: BIDSPath,
            preload: PreloadArg,
    ) -> mne.io.BaseRaw:
        raw = self.source.load(path)
        self.log.info("Raw %s: computing oversampled temporal projection for %s", self.name, path.fpath)
        with user_activity:
            return mne.preprocessing.oversampled_temporal_projection(raw, self.duration)


class RawUpdateBadChannels(CachedRawPipe):

    def __init__(
            self,
            source: str,
            bad_channels: Dict[str, Sequence[str]],
    ):
        CachedRawPipe.__init__(self, source, False)
        self.bad_channels = bad_channels
        self._pattern_keys = [key for key in bad_channels if ('*' in key or '?' in key)]

    def _as_dict(self, args: Sequence[str] = ()) -> dict:
        return CachedRawPipe._as_dict(self, [*args, 'bad_channels'])

    def _make(
            self,
            path: BIDSPath,
            preload: PreloadArg,
    ) -> mne.io.BaseRaw:
        return self.source.load(path, preload=preload)

    def load_bad_channels(
            self,
            path: BIDSPath,
            existing: Collection[str] = None,
    ) -> list[str]:
        bad_channels = self.source.load_bad_channels(path, existing)
        subject = path.entities['subject']
        if subject in self.bad_channels:
            key = subject
        else:
            for key in self._pattern_keys:
                if fnmatch.fnmatch(subject, key):
                    break
            else:
                raise KeyError(subject)
        return sorted(set(bad_channels).union(self.bad_channels[key]))


class RawReReference(CachedRawPipe):
    """Re-reference EEG data

    Parameters
    ----------
    source
        Name of the raw pipe to use for input data.
    reference
        New reference: ``'average'`` (default) or one or several electrode
        names.
    add
        Reconstruct reference channels with given names and set them to 0.
    drop
        Drop these channels after applying the reference.
    cache
        Cache the resulting raw files (default ``False``).

    See Also
    --------
    MneExperiment.raw
    """
    def __init__(
            self,
            source: str,
            reference: Union[str, Sequence[str]] = 'average',
            add: Union[str, Sequence[str]] = None,
            drop: Union[str, Sequence[str]] = None,
            cache: bool = False,
    ):
        CachedRawPipe.__init__(self, source, cache)
        if not isinstance(reference, str):
            reference = tuple_arg('reference', reference, allow_none=False)
        self.reference = reference
        self.add = tuple_arg('add', add)
        self.drop = tuple_arg('drop', drop)

    def _as_dict(self, args: Sequence[str] = ()) -> dict:
        out = CachedRawPipe._as_dict(self, [*args, 'reference'])
        if self.add:
            out['add'] = self.add
        if self.drop:
            out['drop'] = self.drop
        return out

    @staticmethod
    def _normalize_dict(state: dict) -> None:
        if not isinstance(state['reference'], str):
            state['reference'] = tuple_arg('reference', state['reference'])
        for key in ['add', 'drop']:
            if key in state:
                state[key] = tuple_arg(key, state[key])

    def _make(
            self,
            path: BIDSPath,
            preload: PreloadArg,
    ) -> mne.io.BaseRaw:
        raw = self.source.load(path, preload=True)
        if self.add:
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', 'The locations of multiple reference channels are ignored', module='mne')
                raw = mne.add_reference_channels(raw, self.add, copy=False)
            # apply new channel position
            pipe = self.source
            while not isinstance(pipe, RawSource):
                pipe = pipe.source
            if pipe.montage:
                raw.set_montage(pipe.montage)
        raw.set_eeg_reference(list(self.reference))
        if self.drop:
            raw = raw.drop_channels(self.drop)
        return raw

    def _make_info(self, path: BIDSPath) -> mne.io.BaseRaw:
        if self.add or self.drop:
            return self._make(path, False)
        else:
            return super()._make_info(path)


def assemble_pipeline(
        raw: dict[str, RawPipe],
        tasks: tuple[str],
        cache_dir: str,
        deriv_dir: str,
        log: logging.Logger,
) -> dict[str, RawPipe]:
    "Assemble preprocessing pipeline form a definition in a dict"
    linked_raw = {}
    while raw:
        n = len(raw)
        for key in list(raw):
            if raw[key]._can_link(linked_raw):
                pipe = raw.pop(key)._link(key, linked_raw, cache_dir, deriv_dir, log)
                if isinstance(pipe, RawICA):
                    missing = set(pipe.task).difference(tasks)
                    if missing:
                        raise DefinitionError(f"RawICA {key!r} lists one or more non-exising tasks: {', '.join(missing)}")
                linked_raw[key] = pipe
        if len(raw) == n:
            raise DefinitionError(f"Unable to resolve source for raw {enumeration(raw)}, circular dependency?")
    return linked_raw


###############################################################################
# Comparing pipelines
######################


def compare_pipelines(
        old: Dict[str, Dict],
        new: Dict[str, Dict],
        log: logging.Logger,
) -> Tuple[Dict[str, str], Dict[str, str]]:
    """Return a tuple of raw keys for which definitions changed

    Parameters
    ----------
    old
        A {name: params} dict for the previous preprocessing pipeline.
    new
        Current pipeline.
    log
        Logger for logging changes.

    Returns
    -------
    bad_raw : {str: str}
        ``{pipe_name: status}`` dictionary. Status can be 'new', 'removed' or
        'changed'.
    bad_ica : {str: str}
        Same as ``bad_raw`` but only for RawICA pipes (for which ICA files
        might have to be removed).
    """
    out = {}  # status:  good, changed, new, removed
    to_check = []  # need to check whether source is still valid
    keys = set(new).union(old)
    for key in keys:
        new_dict = new.get(key)
        old_dict = old.get(key)
        if new_dict is None:
            out[key] = 'removed'
        elif old_dict is None:
            out[key] = 'new'
        elif new_dict == old_dict:
            if key == 'raw':
                out[key] = 'good'
            else:
                to_check.append(key)
            continue
        else:
            out[key] = 'changed'
        log_dict_change(log, 'raw', key, old_dict, new_dict)

    # secondary changes
    while to_check:
        n = len(to_check)
        for key in tuple(to_check):
            parents = [new[key][k] for k in ('source', 'ica_source') if k in new[key]]
            if any(p not in out for p in parents):
                continue
            elif all(out[p] == 'good' for p in parents):
                out[key] = 'good'
            else:
                out[key] = 'changed'
                log.warning(f"  raw {key} parent changed")
            to_check.remove(key)
        if len(to_check) == n:
            raise RuntimeError("Queue not decreasing")

    bad_raw = {k: v for k, v in out.items() if v != 'good'}
    bad_ica = {k: v for k, v in bad_raw.items() if new.get(k, old.get(k))['type'] == 'RawICA'}
    return bad_raw, bad_ica


def ask_to_delete_ica_files(
        raw: mne.io.BaseRaw,
        status: str,
        filenames: list[str],
) -> None:
    "Ask whether outdated ICA files should be removed and act accordingly"
    if status == 'new':
        msg = ("The definition for raw=%r has been added, but ICA-files "
               "already exist. These files might not correspond to the new "
               "settings and should probably be deleted." % (raw,))
    elif status == 'removed':
        msg = ("The definition for raw=%r has been removed. The corresponsing "
               "ICA files should probably be deleted:" % (raw,))
    elif status == 'changed':
        msg = ("The definition for raw=%r has changed. The corresponding ICA "
               "files should probably be deleted." % (raw,))
    else:
        raise RuntimeError("status=%r" % (status,))
    command = ask(
        "%s Delete %i files?" % (msg, len(filenames)),
        (('abort', 'abort to fix the raw definition and try again'),
         ('delete', 'delete the invalid files'),
         ('ignore', 'pretend that the files are valid; you will not be warned again')))

    if command == 'delete':
        for filename in filenames:
            remove(filename)
    elif command == 'abort':
        raise RuntimeError("User abort")
    elif command != 'ignore':
        raise RuntimeError("command=%r" % (command,))


def normalize_dict(raw: dict) -> None:
    "Normalize pipeline state with latest pipeline classes"
    for key, params in raw.items():
        pipe_class = globals()[params['type']]
        pipe_class._normalize_dict(params)


def remove_task(fname: str) -> str:
    parts = fname.split('_')
    filtered_parts = [part for part in parts if not part.startswith('task-')]
    return '_'.join(filtered_parts)


def remove_subject(fname: str) -> str:
    parts = fname.split('_')
    filtered_parts = [part for part in parts if not part.startswith('sub-')]
    return '_'.join(filtered_parts)


def get_subject_session(fname: str) -> str:
    parts = fname.split('_')
    filtered_parts = [part for part in parts if (part.startswith('sub-') or part.startswith('ses-'))]
    return '_'.join(filtered_parts)



# skip test: data unavailable
from eelbrain import *

from mouse import e


res = e.load_test('surprise', 0.3, 0.5, 0.05, data='sensor', baseline=False, epoch='target', make=True, raw='fastica1-40')
plot.TopoButterfly(res)
# plot.TopoArray(res)
# plot.TopomapBins(res)  # fixme: layout



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Plot multidimensional uniform time series."""
from typing import Any, Union, Sequence, Tuple

import matplotlib.axes
import numpy as np

from .._data_obj import NDVarArg, CategorialArg, IndexArg, Datalist, Dataset
from .._stats.testnd import NDTest
from .._names import INTERPOLATE_CHANNELS
from .._utils import deprecate_ds_arg
from . import _base
from ._base import (
    PlotType, CMapArg,
    EelFigure, PlotData, AxisData, DataLayer, Layout,
    ColorMapMixin, LegendMixin, TimeSlicerEF, TopoMapKey, YLimMixin, XAxisMixin,
    pop_if_dict, set_dict_arg)


class PltIm:

    _aspect = 'auto'

    def __init__(
            self,
            ax: matplotlib.axes.Axes,
            layer: DataLayer,
            cmaps: dict,
            vlims: dict,
            contours: dict,  # {meas: kwargs}
            extent, interpolation, mask=None):
        self.ax = ax
        self._meas = layer.y.info.get('meas')
        self._contours = layer.contour_plot_args(contours)  # kwargs
        self._data = self._data_from_ndvar(layer.y)
        self._extent = extent
        self._mask = mask

        if layer.plot_type == PlotType.IMAGE:
            kwargs = layer.im_plot_args(vlims, cmaps)
            self.im = ax.imshow(self._data, origin='lower', aspect=self._aspect, extent=extent, interpolation=interpolation, **kwargs)
            if mask is not None:
                self.im.set_clip_path(mask)
            self._cmap = kwargs['cmap']
            self.vmin, self.vmax = self.im.get_clim()
        elif layer.plot_type == PlotType.CONTOUR:
            self.im = None
            self.vmin = self.vmax = None
        else:
            raise RuntimeError(f"layer of type {layer.plot_type}")

        # draw flexible parts
        self._contour_h = None
        self._draw_contours()

    def _data_from_ndvar(self, ndvar):
        raise NotImplementedError

    def _draw_contours(self):
        if self._contour_h:
            self._contour_h.remove()
            self._contour_h = None

        if not self._contours:
            return

        # check whether any contours are in data range
        vmin = self._data.min()
        vmax = self._data.max()
        if not any(vmax >= l >= vmin for l in self._contours['levels']):
            return

        self._contour_h = self.ax.contour(self._data, origin='lower', extent=self._extent, **self._contours)
        if self._mask is not None:
            self._contour_h.set_clip_path(self._mask)

    def add_contour(self, meas, level, color):
        if self._meas == meas:
            self._contours['levels'] = [*self._contours.get('levels', ()), level]
            self._contours['colors'] = [*self._contours.get('colors', ()), color]
            self._draw_contours()

    def get_kwargs(self):
        "Get the arguments required for plotting the im"
        if self.im:
            vmin, vmax = self.im.get_clim()
            args = dict(cmap=self._cmap, vmin=vmin, vmax=vmax)
        else:
            args = {}
        return args

    def set_cmap(self, cmap, meas=None):
        if (self.im is not None) and (meas is None or meas == self._meas):
            self.im.set_cmap(cmap)
            self._cmap = cmap

    def set_data(self, ndvar, vlim=False):
        data = self._data_from_ndvar(ndvar)
        if self.im is not None:
            self.im.set_data(data)
            if vlim:
                vmin, vmax = _base.find_vlim_args(ndvar)
                self.set_vlim(vmin, vmax, None)

        self._data = data
        self._draw_contours()

    def set_vlim(self, v, vmax=None, meas=None):
        if self.im is None:
            return
        elif (meas is not None) and (self._meas != meas):
            return
        vmin, vmax = _base.fix_vlim_for_cmap(v, vmax, self._cmap)
        self.im.set_clim(vmin, vmax)
        self.vmin, self.vmax = self.im.get_clim()


class PltImArray(PltIm):

    def __init__(self, ax, layer, dimnames, interpolation, vlims, cmaps, contours):
        self._dimnames = dimnames[::-1]
        xdim, ydim = layer.y.get_dims(dimnames)
        xlim = xdim._axis_im_extent()
        ylim = ydim._axis_im_extent()
        PltIm.__init__(self, ax, layer, cmaps, vlims, contours, xlim + ylim, interpolation)
        ax.set_xlim(xlim)
        ax.set_ylim(ylim)

    def _data_from_ndvar(self, ndvar):
        return ndvar.get_data(self._dimnames)


class AxImArray:

    def __init__(self, ax, layers, x='time', interpolation=None, vlims={}, cmaps={}, contours={}):
        self.ax = ax
        self.data = layers
        dimnames = layers.y0.get_dimnames((x, None))
        self.plots = [PltImArray(ax, l, dimnames, interpolation, vlims, cmaps, contours) for l in layers]

    @property
    def title(self):
        return self.ax.get_title()

    def add_contour(self, meas, level, color):
        for l in self.plots:
            l.add_contour(meas, level, color)

    def set_cmap(self, cmap, meas=None):
        """Change the colormap in the array plot

        Parameters
        ----------
        cmap : str | colormap
            New colormap.
        meas : None | str
            Measurement to which to apply the colormap. With None, it is
            applied to all.
        """
        for l in self.plots:
            l.set_cmap(cmap, meas)

    def set_data(self, layers, vlim=False):
        """Update the plotted data

        Parameters
        ----------
        layers : list of NDVar
            Data to plot
        vlim : bool
            Update vlims for the new data.
        """
        for l, p in zip(layers, self.plots):
            p.set_data(l, vlim)

    def set_vlim(self, v, vmax=None, meas=None):
        for l in self.plots:
            l.set_vlim(v, vmax, meas)


class Array(TimeSlicerEF, ColorMapMixin, XAxisMixin, EelFigure):
    """Plot UTS data to a rectangular grid.

    Parameters
    ----------
    y
        Data to plot.
    xax
        Create a separate plot for each cell in this model.
    xlabel
        Labels for x-axis; the default is determined from the data.
    ylabel
        Labels for y-axis; the default is determined from the data.
    xticklabels
        Specify which axes should be annotated with x-axis tick labels.
        Use ``int`` for a single axis, a sequence of ``int`` for multiple
        specific axes, or one of ``'left' | 'bottom' | 'all' | 'none'``.
    yticklabels
        Specify which axes should be annotated with y-axis tick labels.
        Use ``int`` for a single axis, a sequence of ``int`` for multiple
        specific axes, or one of ``'left' | 'bottom' | 'all' | 'none'``.
    data : Dataset
        If a Dataset is provided, ``epochs`` and ``xax`` can be specified
        as strings.
    sub
        Specify a subset of the data.
    x
        Dimension to plot on the x axis (default 'time').
    vmax
        Upper limits for the colormap.
    vmin
        Lower limit for the colormap.
    cmap
        Colormap (default depends on the data).
    contours
        Draw contours. Can be an int (number of contours, including
        ``vmin`` and ``vmax``), a sequence (values at which to draw
        contours), or a dictionary with ``**kwargs`` for
        :meth:`~matplotlib.axes.Axes.contour` (must include a ``"levels"`` key).
        Default is no contours.
    axtitle
        Title for the individual axes. The default is to show the names of the
        epochs, but only if multiple axes are plotted.
    interpolation
        Array image interpolation (see Matplotlib's
        :meth:`~matplotlib.axes.Axes.imshow`). Matplotlib 1.5.3's SVG output
        can't handle uneven aspect with ``interpolation='none'``, use
        ``interpolation='nearest'`` instead.
    xlim
        Initial x-axis view limits as ``(left, right)`` tuple or as ``length``
        scalar (default is the full x-axis in the data).
    tight : bool
        Use matplotlib's tight_layout to expand all axes to fill the figure
        (default True)
    ...
        Also accepts :ref:`general-layout-parameters`.

    Notes
    -----
    Navigation:
     - ``←``: scroll left
     - ``→``: scroll right
     - ``home``: scroll to beginning
     - ``end``: scroll to end
     - ``f``: zoom in (reduce x axis range)
     - ``d``: zoom out (increase x axis range)
    """
    @deprecate_ds_arg
    def __init__(
            self,
            y: Union[NDVarArg, Sequence[NDVarArg]],
            xax: CategorialArg = None,
            xlabel: Union[bool, str] = True,
            ylabel: Union[bool, str] = True,
            xticklabels: Union[str, int, Sequence[int]] = 'bottom',
            yticklabels: Union[str, int, Sequence[int]] = 'left',
            sub: IndexArg = None,
            data: Dataset = None,
            x: str = 'time',
            vmax: float = None,
            vmin: float = None,
            cmap: CMapArg = None,
            contours: Union[int, Sequence, dict] = None,
            axtitle: Union[bool, Sequence[str]] = True,
            interpolation: str = None,
            xlim: Union[float, Tuple[float, float]] = None,
            **kwargs):
        plot_data = PlotData.from_args(y, (x, None), xax, data, sub).for_plot(PlotType.IMAGE)
        xdim, ydim = plot_data.dims
        self.plots = []
        ColorMapMixin.__init__(self, plot_data.data, cmap, vmax, vmin, contours, self.plots)

        layout = Layout(plot_data.plot_used, 1.5, 3, **kwargs)
        EelFigure.__init__(self, plot_data.frame_title, layout)
        self._set_axtitle(axtitle, plot_data)

        for ax, layers in zip(self.axes, plot_data):
            p = AxImArray(ax, layers, x, interpolation, self._vlims, self._cmaps, self._contours)
            self.plots.append(p)

        self._configure_axis_dim('x', xdim, xlabel, xticklabels, data=plot_data.data)
        self._configure_axis_dim('y', ydim, ylabel, yticklabels, scalar=False, data=plot_data.data)
        XAxisMixin._init_with_data(self, plot_data.data, xdim, xlim, im=True)
        TimeSlicerEF.__init__(self, xdim, plot_data.time_dim)
        self._show()

    def _fill_toolbar(self, tb):
        ColorMapMixin._fill_toolbar(self, tb)


class PltUTSND:

    def __init__(
            self,
            ax: matplotlib.axes.Axes,
            layer: DataLayer,
            xdim: str,
            line_dim: str,
            sensors: Union[Sequence, bool] = None,
            **kwargs,
    ):
        epoch = layer.y
        if sensors is not None and sensors is not True:
            epoch = epoch.sub(sensor=sensors)

        kwargs = layer.plot_args(kwargs)
        color = pop_if_dict(kwargs, 'color')
        z_order = pop_if_dict(kwargs, 'zorder')
        self._dims = (line_dim, xdim)
        x = epoch.get_dim(xdim)._axis_data()
        y = epoch.get_data((xdim, line_dim))
        line_dim_obj = epoch.get_dim(line_dim)
        self.legend_handles = {}
        self.lines = ax.plot(x, y, label=epoch.name, **kwargs)

        # apply line-specific formatting
        lines = Datalist(self.lines)
        if z_order:
            set_dict_arg('zorder', z_order, line_dim_obj, lines)

        if color:
            self.legend_handles = {}
            set_dict_arg('color', color, line_dim_obj, lines, self.legend_handles)
        else:
            self.legend_handles = {epoch.name: self.lines[0]}

        for y, kwa in _base.find_uts_hlines(epoch):
            ax.axhline(y, **kwa)

        self.epoch = epoch
        self._sensors = sensors

    def remove(self):
        while self.lines:
            self.lines.pop().remove()

    def set_visible(self, visible=True):
        for line in self.lines:
            line.set_visible(visible)

    def set_ydata(self, epoch):
        if self._sensors:
            epoch = epoch.sub(sensor=self._sensors)
        for line, y in zip(self.lines, epoch.get_data(self._dims)):
            line.set_ydata(y)


class AxButterfly:
    def __init__(
            self,
            ax: matplotlib.axes.Axes,
            axis_data: AxisData,
            xdim: str,
            linedim: str,
            sensors,
            color,
            linewidth,
            vlims,
            clip: bool = True,
    ):
        self.ax = ax
        self.data = axis_data.ndvars
        self.layers = []
        self.legend_handles = {}
        self._meas = None

        vmin, vmax = _base.find_uts_ax_vlim(axis_data.ndvars, vlims)

        name = ''
        for layer in axis_data:
            h = PltUTSND(ax, layer, xdim, linedim, sensors, clip_on=clip, color=color, linewidth=linewidth)
            self.layers.append(h)
            if not name and layer.y.name:
                name = layer.y.name

            self.legend_handles.update(h.legend_handles)

        ax.yaxis.offsetText.set_va('top')

        self.set_ylim(vmin, vmax)

    @property
    def title(self):
        return self.ax.get_title()

    def set_ylim(self, vmin, vmax):
        self.ax.set_ylim(vmin, vmax)
        self.vmin, self.vmax = self.ax.get_ylim()


class Butterfly(TimeSlicerEF, LegendMixin, TopoMapKey, YLimMixin, XAxisMixin, EelFigure):
    """Butterfly plot for NDVars

    Parameters
    ----------
    y : (list of) NDVar
        Data to plot.
    xax : None | categorial
        Create a separate plot for each cell in this model.
    sensors: None or list of sensor IDs
        sensors to plot (``None`` = all)
    axtitle : bool | sequence of str
        Title for the individual axes. The default is to show the names of the
        epochs, but only if multiple axes are plotted.
    xlabel
        X-axis label. By default the label is inferred from the data.
    ylabel
        Y-axis label. By default the label is inferred from the data.
    xticklabels
        Specify which axes should be annotated with x-axis tick labels.
        Use ``int`` for a single axis, a sequence of ``int`` for multiple
        specific axes, or one of ``'left' | 'bottom' | 'all' | 'none'``.
    yticklabels
        Specify which axes should be annotated with y-axis tick labels.
        Use ``int`` for a single axis, a sequence of ``int`` for multiple
        specific axes, or one of ``'left' | 'bottom' | 'all' | 'none'``.
    color : matplotlib color | dict
        Either a color for all lines, or a dictionary mapping levels of the
        line dimension to colors. The default is to use ``NDVar.info['color']``
        if available, otherwise the matplotlib default color alternation. Use
        ``color=True`` to use the matplotlib default.
    linewidth : scalar
        Linewidth for plots (defult is to use ``matplotlib.rcParams``).
    data : Dataset
        If a Dataset is provided, ``epochs`` and ``xax`` can be specified
        as strings.
    sub : str | array
        Specify a subset of the data.
    x : str
        Dimension to plot on the x-axis (default 'time').
    vmax : scalar
        Top of the y axis (default depends on data).
    vmin : scalar
        Bottom of the y axis (default depends on data).
    xlim : scalar | (scalar, scalar)
        Initial x-axis view limits as ``(left, right)`` tuple or as ``length``
        scalar (default is the full x-axis in the data).
    clip : bool
        Clip lines outside of axes (the default depends on whether ``frame`` is
        closed or open).
    tight : bool
        Use matplotlib's tight_layout to expand all axes to fill the figure
        (default True)
    ...
        Also accepts :ref:`general-layout-parameters`.

    Notes
    -----
    Navigation:
     - ``↑``: scroll up
     - ``↓``: scroll down
     - ``←``: scroll left
     - ``→``: scroll right
     - ``home``: scroll to beginning
     - ``end``: scroll to end
     - ``f``: x-axis zoom in (reduce x axis range)
     - ``d``: x-axis zoom out (increase x axis range)
     - ``r``: y-axis zoom in (reduce y-axis range)
     - ``c``: y-axis zoom out (increase y-axis range)

    Keys available for sensor data:
     - ``t``: open a ``Topomap`` plot for the time point under the mouse pointer.
     - ``T``: open a larger ``Topomap`` plot with visible sensor names for the
       time point under the mouse pointer.
    """
    _cmaps = None  # for TopoMapKey mixin
    _contours = None
    # keep track of open butterfly combo plots to optimally position new plots on screen
    _OPEN_PLOTS = []

    @deprecate_ds_arg
    def __init__(
            self,
            y: Union[NDVarArg, Sequence, NDTest],
            xax: CategorialArg = None,
            sensors: Sequence = None,
            axtitle: Union[bool, Sequence[str]] = True,
            xlabel: Union[bool, str] = True,
            ylabel: Union[bool, str] = True,
            xticklabels: Union[str, int, Sequence[int]] = 'bottom',
            yticklabels: Union[str, int, Sequence[int]] = 'left',
            color: Any = None,
            linewidth: float = None,
            data: Dataset = None,
            sub: IndexArg = None,
            x: str = 'time',
            vmax: float = None,
            vmin: float = None,
            xlim: Union[float, Tuple[float, float]] = None,
            clip: bool = None,
            **kwargs,
    ):
        plot_data = PlotData.from_args(y, (x, None), xax, data, sub).for_plot(PlotType.LINE)
        xdim, linedim = plot_data.dims
        layout = Layout(plot_data.plot_used, 2, 4, **kwargs)
        EelFigure.__init__(self, plot_data.frame_title, layout)
        self._set_axtitle(axtitle, plot_data)
        self._configure_axis_dim('x', xdim, xlabel, xticklabels, data=plot_data.data)
        self._configure_axis_data('y', plot_data.y0, ylabel, yticklabels)

        if clip is None:
            clip = layout.frame is True

        self.plots = []
        self._vlims = _base.find_fig_vlims(plot_data.data, vmax, vmin)
        legend_handles = {}
        for ax, layers in zip(self.axes, plot_data):
            h = AxButterfly(ax, layers, xdim, linedim, sensors, color, linewidth, self._vlims, clip)
            self.plots.append(h)
            legend_handles.update(h.legend_handles)

        XAxisMixin._init_with_data(self, plot_data.data, xdim, xlim)
        YLimMixin.__init__(self, self.plots)
        if linedim == 'sensor':
            TopoMapKey.__init__(self, self._topo_data)
        LegendMixin.__init__(self, 'invisible', legend_handles)
        TimeSlicerEF.__init__(self, xdim, plot_data.time_dim)
        self._show()

    def _auto_position(self, p2=None):
        """Position Butterfly plot and corresponding time-slice plot"""
        from .._wxgui import wx
        from .._wxgui.mpl_canvas import CanvasFrame

        if not isinstance(self._frame, CanvasFrame):
            return
        px, py = wx.ClientDisplayRect()[:2]
        display_w, display_h = wx.DisplaySize()
        pw, ph = self._frame.GetSize()
        old_ys = []
        self.__class__._OPEN_PLOTS = [p for p in self._OPEN_PLOTS if p._frame is not None]
        if self._OPEN_PLOTS:
            for p in self._OPEN_PLOTS:
                old_x, old_y = p._frame.GetPosition()
                if old_x < px + 10:
                    _, old_h = p._frame.GetSize()
                    old_ys.append((old_y, old_y + old_h))
            if old_ys:
                old_ys.sort()
                # find vertical space without overlap
                for y_start, y_stop in old_ys:
                    overlap = max(0, min(py + ph, y_stop) - max(py, y_start))
                    # print(f"overlap: [{y_start}, {y_stop}], [{py}, {py + ph}] -> {overlap}")
                    if overlap:
                        py = y_stop
                    else:
                        break
                py = min(y_stop, display_h - ph)

        self._frame.SetPosition((px, py))
        self._OPEN_PLOTS.append(self)

        if p2:  # secondary plot
            p2w, _ = p2._frame.GetSize()
            p2x = min(px + pw, display_w - p2w)
            p2._frame.SetPosition((p2x, py))

    def _fill_toolbar(self, tb):
        LegendMixin._fill_toolbar(self, tb)

    def _topo_data(self, event):
        if not event.inaxes:
            return
        p = self.plots[self.axes.index(event.inaxes)]
        t = event.xdata
        data = [l.sub(time=t) for l in p.data]
        return data, p.title + ' %i ms' % round(t), 'default'


class AxButterflyEpoch:

    def __init__(self, ax, epoch, mark=None, state=True, label=None, color='k',
                 lw=0.2, mcolor='r', mlw=0.8, antialiased=True, vlims={}):
        """Specific plot for showing a single sensor by time epoch

        Parameters
        ----------
        ax : mpl Axes
            Plot target axes.
        epoch : NDVar
            Sensor by time epoch.
        mark : dict {int: mpl color}
            Channel: color dict of channels with custom color.
        color : mpl color
            Color for unmarked traces.
        lw : scalar
            Sensor trace plot Line width (default 0.5).
        mlw : scalar
            Marked sensor plot line width (default 1).
        """
        self.lines = PltUTSND(ax, DataLayer(epoch, PlotType.LINE), 'time', 'sensor',
                              color=color, lw=lw, antialiased=antialiased)
        ax.set_xlim(epoch.time[0], epoch.time[-1])

        self.ax = ax
        self.epoch = epoch
        self._state_h = []
        self._visible = True
        self.set_ylim(_base.find_uts_ax_vlim([epoch], vlims))
        self._styles = {None: {'color': color, 'lw': lw, 'ls': '-',
                               'zorder': 2},
                        'mark': {'color': mcolor, 'lw': mlw, 'ls': '-',
                                 'zorder': 10},
                        INTERPOLATE_CHANNELS: {'color': 'b', 'lw': 1.2,
                                               'ls': ':', 'zorder': 6}}
        self._marked = {'mark': set(), INTERPOLATE_CHANNELS: set()}
        if mark:
            self.set_marked('mark', mark)

        if label is None:
            label = ''
        self._label = ax.text(0, 1.01, label, va='bottom', ha='left', transform=ax.transAxes)

        # create initial plots
        self.set_state(state)

    def set_data(self, epoch, label=None):
        self.epoch = epoch
        self.lines.set_ydata(epoch)
        if label is not None:
            self._label.set_text(label)

    def set_marked(self, kind, sensors):
        """Set the channels which should be marked for a specific style

        Parameters
        ----------
        kind : str
            The style.
        sensors : collection of int
            Channel index for the channels to mark as ``kind``.
        """
        old = self._marked[kind]
        new = self._marked[kind] = set(sensors)
        if not old and not new:
            return
        # mark new channels
        for i in new.difference(old):
            self.lines.lines[i].update(self._styles[kind])
        # find channels to unmark
        old.difference_update(new)
        if not old:
            return
        # possible alternate style
        if kind == 'mark':
            other_kind = INTERPOLATE_CHANNELS
        else:
            other_kind = 'mark'
        # reset old channels
        for i in old:
            if i in self._marked[other_kind]:
                self.lines.lines[i].update(self._styles[other_kind])
            else:
                self.lines.lines[i].update(self._styles[None])

    def set_state(self, state):
        "Set the state (True=accept / False=reject)"
        if state:
            while self._state_h:
                h = self._state_h.pop()
                h.remove()
        else:
            if not self._state_h:
                h1 = self.ax.plot([0, 1], [0, 1], color='r', linewidth=1,
                                  transform=self.ax.transAxes)
                h2 = self.ax.plot([0, 1], [1, 0], color='r', linewidth=1,
                                  transform=self.ax.transAxes)
                self._state_h.extend(h1 + h2)

    def set_visible(self, visible=True):
        if self._visible != visible:
            self.lines.set_visible(visible)
            self._label.set_visible(visible)
            for line in self._state_h:
                line.set_visible(visible)
            self._visible = visible

    def set_ylim(self, ylim):
        if ylim:
            if np.isscalar(ylim):
                self.ax.set_ylim(-ylim, ylim)
            else:
                self.ax.set_ylim(ylim)



"""
.. _exa-mtrf:

EEG speech envelope TRF
=======================
A TRF analysis on data that comes as arrays.
Analyze continuous speech data from the mTRF dataset [1]_.
Use the boosting algorithm for estimating temporal response functions (TRFs)
to the acoustic envelope.

.. contents:: Sections
    :local:
    :backlinks: top

"""
# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
# sphinx_gallery_thumbnail_number = 4
import os

from scipy.io import loadmat
import mne
from eelbrain import *

# Load the mTRF speech dataset and convert data to NDVars
root = mne.datasets.mtrf.data_path()
speech_path = os.path.join(root, 'speech_data.mat')
mdata = loadmat(speech_path)

# Time axis
tstep = 1. / mdata['Fs'][0, 0]
n_times = mdata['envelope'].shape[0]
time = UTS(0, tstep, n_times)
# Load the EEG sensor coordinates (drop fiducials coordinates, which are stored
# after sensor 128)
sensor = Sensor.from_montage('biosemi128')[:128]
# Frequency dimension for the spectrogram
band = Scalar('frequency', range(16))
# Create variables
envelope = NDVar(mdata['envelope'][:, 0], (time,), name='envelope')
eeg = NDVar(mdata['EEG'], (time, sensor), name='EEG', info={'unit': 'µV'})
spectrogram = NDVar(mdata['spectrogram'], (time, band), name='spectrogram')
# Exclude a bad channel
eeg = eeg[sensor.index(exclude='A13')]

###############################################################################
# Data
# ----
# Plot the spectrogram of the speech stimulus:
plot.Array(spectrogram, xlim=5, w=6, h=2)

###############################################################################
# Plot the envelope used as stimulus representation for TRFs:
plot.UTS(envelope, xlim=5, w=6, h=2)

###############################################################################
# Plot the corresponding EEG data:
p = plot.TopoButterfly(eeg, xlim=5, w=7, h=2)
p.set_time(1.200)

###############################################################################
# TRF estimation
# --------------
# TRF for the envelope using boosting:
#
# - TRF from -100 to 400 ms
# - Basis of 100 ms Hamming windows
# - Use 4 partitionings of the data for cross-validation based early stopping
res = boosting(eeg, envelope, -0.100, 0.400, basis=0.100, partitions=4)
p = plot.TopoButterfly(res.h_scaled, w=6, h=2)
p.set_time(.180)

###############################################################################
# Multiple predictors
# -------------------
# Multiple predictors additively explain the signal:

# Derive acoustic onsets from the envelope
onset = envelope.diff('time', name='onset').clip(0)
onset *= envelope.max() / onset.max()
plot.UTS([[envelope, onset]], xlim=5, w=6, h=2)

###############################################################################
res_onset = boosting(eeg, [onset, envelope], -0.100, 0.400, basis=0.100, partitions=4)
p = plot.TopoButterfly(res_onset.h_scaled, w=6, h=3)
p.set_time(.150)

###############################################################################
# Compare models
# --------------
# Compare model quality through the correlation between measured and predicted
# responses:
plot.Topomap([res.r, res_onset.r], w=4, h=2, columns=2, axtitle=['envelope', 'envelope + onset'])


###############################################################################
# References
# ----------
# .. [1] Crosse, M. J., Liberto, D., M, G., Bednar, A., & Lalor, E. C. (2016).
#        The Multivariate Temporal Response Function (mTRF) Toolbox: A MATLAB
#        Toolbox for Relating Neural Signals to Continuous Stimuli. Frontiers in
#        Human Neuroscience, 10. https://doi.org/10.3389/fnhum.2016.00604



"""
.. currentmodule:: eelbrain


Group level analysis
====================

:class:`Datasets <Dataset>` provide a means to collect data for statistical
analysis.
A :class:`Dataset` is similar to a dataframe in R or :mod:`pandas`,
but can hold mass-univariate measurements in :class:`NDVars <NDVar>`.

This example illustrates how to construct a :class:`Datasets <Dataset>`
by first collecting the cases, or rows, of the desired data table,
and then combining them using :meth:`Dataset.from_caselist`.

.. contents:: Contents
   :local:

"""

###############################################################################
# Simulating subject data
# -----------------------
#
# This example uses simulated EEG data to illustrate that.
#
# We will use the simulated EEG data to derive a general input structure for
# creating group level datasets:
# A list of condition labels and corresponding (simulated) EEG responses.

from eelbrain import *


data = datasets.simulate_erp(seed=1)
data.head()

###############################################################################
# Average the data by condition to get two condition averages per subject:

data_by_condition = data.aggregate('predictability')
data_by_condition

###############################################################################
# Turn this into the general label/brain response structure:

subject_data = list(data_by_condition.zip('predictability', 'eeg'))
subject_data

###############################################################################
# Construct group level data
# --------------------------
#
# Use the procedure described above to simulate a group level dataset.
#
# We collect the labels (subject and condition labels) and brain responses in
# a list (``cases``).
# Each entry in this list corresponds to one row of the desired :class:`Dataset`:

cases = []  # list of rows
for subject in range(10):
    data = datasets.simulate_erp(seed=subject)
    data_by_condition = data.aggregate('predictability')
    for predictability, eeg in data_by_condition.zip('predictability', 'eeg'):
        cases.append([str(subject), predictability, eeg])

cases

###############################################################################
# This list can now be turned into a :class:`Dataset`:

data = Dataset.from_caselist(['subject', 'predictability', 'eeg'], cases, random='subject')
data.head()

###############################################################################
# Averaging by condition
# ----------------------
#
# In a dataset that contains condition labels, these labels can be used to derive averages by condition:

data_by_condition = data.aggregate('predictability', drop_bad=True)
data_by_condition

###############################################################################
# This could be used to retrieve those average responses:

data_by_condition[0, 'eeg']

###############################################################################
# The grand average could be derived by aggregating without a model, resulting
# in a single row:

data.aggregate(drop_bad=True)

###############################################################################
# Many functions automatically average across cases ...

p = plot.TopoButterfly('eeg', data=data)

###############################################################################
# ... and directly accept a parameter for averaging by condition:

p = plot.TopoButterfly('eeg', 'predictability', data=data)

###############################################################################
# Models can similarly be used to define conditions in statistical tests:

result = testnd.TTestRelated('eeg', 'predictability', match='subject', data=data)
result



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""
Permutation statistics
======================

.. currentmodule:: eelbrain

Eelbrain implents three methods for estimating null-distributions in permutation tests:
max-statistic, cluster-mass and TFCE.
This example illustrates these using a simple *t*-test.
For simulating more complex datasets see other examples in this section.

For the sake of speed, the tests here are based on 1000 permutations of the data (``samples=1000``).
For precise *p*-values, 10000 permutations (the default) are preferable.

.. contents:: Contents
   :local:

"""
# sphinx_gallery_thumbnail_number = 2
from eelbrain import *
import scipy.stats


data = datasets.simulate_erp()
eeg = data['eeg']
p = plot.TopoButterfly(eeg, t=0.13, head_radius=0.35)

###############################################################################
# Max statistic
# -------------
# The max statistic is simply the maximum across the entire *t*-map. 
# The permutation distribution consist of ``max(t)`` for each permutation. 
# This is the default, and is also the fastest test.

result = testnd.TTestOneSample(eeg, samples=1000)

###############################################################################
# Visualize the permutation distribution.
# We consider values significant if they are larger than 95% of the permutation
# distribution, so we look for the 95th percentile:

significance_threshold = scipy.stats.scoreatpercentile(result.permutation_distribution, 95)
p = plot.Histogram(
    result.permutation_distribution, title='Distribution of $max(t)$',
    xlabel='max(t)', ylabel='Number of permutations')
p.add_vline(significance_threshold, color='red', linestyle='--')
_ = p.axes[0].text(significance_threshold, 40, ' 95%', color='red')

###############################################################################
# In the actual comparison, any data point with a *t*-value that is more extreme
# than 95% of the permutation distribution is considered signifiacant:

p = plot.TopoButterfly(result, t=0.13, head_radius=0.35)

###############################################################################
# Significant regions can be retrieved as :class:`Dataset` of :class:`NDVar` for further analysis (e.g., for defining ROIs).
# This function will find all contiguous regions in which *p* ≤ .05:

significan_regions = result.find_clusters(0.05, maps=True)
significan_regions

###############################################################################
p = plot.TopoButterfly(significan_regions[0, 'cluster'], t=0.13, head_radius=0.35)

###############################################################################
# Cluster-based tests
# -------------------
# In cluster-based tests, a first steps consists in finding contiguous regions of "meaningful" effect, so-called clusters.
# In order to find contiguous regions, the algorithm needs to know which channels are
# neighbors. This information is refered to as the sensor adjacency (i.e., which sensors
# are connected). The adjacency graph can be visualized to confirm that it is set correctly.

p = plot.SensorMap(eeg, adjacency=True)

###############################################################################
# We also need to define what constitutes a "meaningful" effect, i.e., an effect that should be included in the cluster.
# This can be defined based on the magnitude of the statistic (e.g., a *t*-value).
# It is commonly set to a *t* value equivalent to an uncorrected *p*-vale of .05.
# This is specified with the ``pmin=.05`` argument.
#
# A summary statistic is then computed for each cluster. In Eelbrain this is the *cluster-mass* statistic:
# the sum of all values in the cluster
# (i.e., the sum of the *t* values at all datapoints that are part of the cluster).
# The clustering procedure is repeated in each permutation, and the largest cluster-mass value is retained, 
# to derive a permutation distribution for cluster-mass values expected under the null hypothesis.
# Each cluster in the actual data can then be evaluated against this distribution.

result = testnd.TTestOneSample(eeg, samples=1000, pmin=.05)
p = plot.TopoButterfly(result, t=0.13, head_radius=0.35)

###############################################################################
# Compared to the ``max(t)`` approach above:
#
# 1) Clusters tend to be larger, because an uncorrected *p* = .05 (the cluster forming threshold) usually corresponds to a lower *t*-value than a corrected *p* = .05.
# 2) A second, later cluster emerged, showing that this test can be more powerful when effects are temporally and spatially extended.
#
# Because clusters are formed before computing the permutation distribution, there are usually many non-significant clusters (we don not want to list them all here):

result.find_clusters().head()

###############################################################################
# List only significant clusters
result.find_clusters(pmin=0.05)

###############################################################################
# Threshold-free cluster enhancement (TFCE)
# -----------------------------------------
# TFCE is an image processing algorithm that enhances cluster-like features, but without setting an arbitrary threshold.
# It thus combines the advantage the max statistic approach (not having to set an arbitrary threshold) with the advantage of the cluster-based test of increased sensitivity for effects that are extended in space and time. 

result = testnd.TTestOneSample(eeg, samples=1000, tfce=True)
p = plot.TopoButterfly(result, t=0.13, head_radius=0.35)

###############################################################################
# Result representation is analogous to the max statistic approach, with the addition that the TFCE map can be visualized:

p = plot.TopoButterfly(result.tfce_map, t=0.13, head_radius=0.35)



# skip test: data unavailable
from os import makedirs, mkdir
from pathlib import Path
from posixpath import dirname
import mne

# Download the raw dataset at https://openneuro.org/datasets/ds006035/versions/1.0.0, then run this script.

MRI_SDIR = '/mnt/d/Data/Somatomotor/derivatives/freesurfer'
RAW_FILE = '/mnt/d/Data/Somatomotor/sub-{subject}/ses-meeg/meg/sub-{subject}_ses-meeg_task-somatomotor_run-1_meg.fif'
TRANS_FILE = '/mnt/d/Data/Somatomotor/derivatives/trans/{subject}_trans.fif'
subjects = ['sm04', 'sm06', 'sm07', 'sm09', 'sm12']

mne.datasets.fetch_fsaverage(subjects_dir=Path(MRI_SDIR))
makedirs(dirname(TRANS_FILE), exist_ok=True)
for subject in subjects:
    mne.scale_mri('fsaverage', subject, 1., subjects_dir=MRI_SDIR, labels=False)
    raw = mne.io.read_raw_fif(RAW_FILE.format(subject=subject), preload=False)
    coreg = mne.coreg.Coregistration(raw.info, subject=subject, subjects_dir=MRI_SDIR)
    coreg.fit_fiducials()
    coreg.fit_icp()
    mne.write_trans(TRANS_FILE.format(subject=subject), coreg.trans, overwrite=True)



# skip test: data unavailable
from os import makedirs
from posixpath import dirname
import mne


MRI_SDIR = '/mnt/d/Data/ds005810/derivatives/freesurfer'
RAW_FILE = '/mnt/d/Data/ds005810/sub-{subject}/ses-ImageNet01/meg/sub-{subject}_ses-{session}_task-ImageNet_run-01_meg.fif'
TRANS_FILE = '/mnt/d/Data/ds005810/derivatives/trans/sub-{subject}_ses-{session}_trans.fif'
subjects = ['01', '03', '04']
sessions = ['ImageNet01']


makedirs(dirname(TRANS_FILE), exist_ok=True)
for subject in subjects:
    for session in sessions:
        raw = mne.io.read_raw_fif(RAW_FILE.format(subject=subject, session=session), preload=False)
        coreg = mne.coreg.Coregistration(raw.info, subject=subject, subjects_dir=MRI_SDIR)
        coreg.fit_fiducials()
        coreg.fit_icp()
        mne.write_trans(TRANS_FILE.format(subject=subject, session=session), coreg.trans, overwrite=True)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Continuous neural data (CND) format used for mTRF-Toolbox"""
import logging
from pathlib import Path
from typing import Optional, Sequence, Union

import mne
import numpy
from pymatreader import read_mat

from .._data_obj import Dataset, Factor, NDVar, Scalar, Sensor, UTS, Var, combine, _matrix_graph
from .._io.fiff import mne_neighbor_files
from .._types import PathArg
from .._utils import deprecate_kwarg, ui


FILETYPES = [("CND files", "*.mat")]


@deprecate_kwarg('connectivity', 'adjacency', '0.41', '0.42')
def read_cnd(
        filename: PathArg = None,
        adjacency: Union[str, Sequence, float] = None,
) -> Optional[Dataset]:
    """Load continuous neural data (CND) file used in the mTRF-Toolbox

    Parameters
    ----------
    filename
        Path to the data file (``*.mat``). If unspecified, open a file dialog.
    adjacency
        Sensor adjacency graph for EEG sensors.
        By default, the function tries to use the ``deviceName`` entry and falls
        back on distance-based adjacency for unkown devices.
        Can be explicitly specified as a `FieldTrip neighbor file
        <https://www.fieldtriptoolbox.org/template/neighbours/>`_ (e.g., ``'biosemi64'``;
        Use a :class:`float` for distance-based adjacency (see :meth:`Sensor.set_adjacency`).
        For more options see :class:`Sensor`.

    Notes
    -----
    This format is experimental and the returned data format might change in the future;
    Please report any problems you might encounter.
    """
    if filename is None:
        path = ui.ask_file("Load CND File", "Select CND file to load as NDVar", FILETYPES)
        if not path:
            return
    else:
        path = Path(filename)
        if not path.suffix and not path.exists():
            path = path.with_suffix('.mat')
    data = read_mat(path)
    if 'eeg' in data:
        data_type = data['eeg']['dataType']
        # EEG sensor properties
        dist_adjacency = None
        sysname = data['eeg']['deviceName']
        ch_names = data['eeg']['chanlocs']['labels']
        # adjacency default
        if adjacency is None:
            available = mne_neighbor_files()
            if sysname in available:
                adjacency = sysname
            else:
                adjacency = 1.6
        # find adjacency
        if isinstance(adjacency, float):
            dist_adjacency = adjacency
            adjacency = 'none'
        elif adjacency is False:
            adjacency = 'none'
        elif isinstance(adjacency, str) and adjacency not in ('grid', 'none'):
            adj_matrix, adj_names = mne.channels.read_ch_adjacency(adjacency)
            # fix channel order
            if adj_names != ch_names:
                index = numpy.array([adj_names.index(name) for name in ch_names])
                adj_matrix = adj_matrix[index][:, index]
            adjacency = _matrix_graph(adj_matrix)
        locs = numpy.vstack([
            -numpy.array(data['eeg']['chanlocs']['Y']),
            data['eeg']['chanlocs']['X'],
            data['eeg']['chanlocs']['Z'],
        ]).T
        sensor = Sensor(locs, ch_names, sysname, adjacency=adjacency)
        if dist_adjacency:
            sensor.set_adjacency(connect_dist=dist_adjacency)
        # EEG data
        tstep = 1 / data['eeg']['fs']
        eeg = []
        for trial_data in data['eeg']['data']:
            uts = UTS(0, tstep, trial_data.shape[0])
            eeg.append(NDVar(trial_data, (uts, sensor), 'eeg'))
        ds = Dataset({data_type.lower(): combine(eeg, to_list=True)})
        # Trial position
        if 'origTrialPosition' in data['eeg']:
            orig_trial_position = data['eeg']['origTrialPosition']
            if len(orig_trial_position) != ds.n_cases:
                logger = logging.getLogger(__name__)
                logger.warning(f"Ignoring origTrialPosition because it has the wrong length: {orig_trial_position!r}")
            else:
                ds['origTrialPosition'] = Var(orig_trial_position - 1)
        else:
            logger = logging.getLogger(__name__)
            logger.warning("origTrialPosition missing")
        # Extra channels
        if 'extChan' in data['eeg']:
            desc = ds.info['extChan'] = data['eeg']['extChan']['description']
            extra_data = []
            for trial_data in data['eeg']['extChan']['data']:
                uts = UTS(0, tstep, trial_data.shape[0])
                channel = Scalar('channel', range(trial_data.shape[1]))
                extra_data.append(NDVar(trial_data, (uts, channel), desc))
            ds['extChan'] = extra_data
        if 'reRef' in data['eeg']:
            ds.info['reRef'] = data['eeg']['reRef']
        return ds
    if 'stim' in data:
        ds = Dataset({'stimIdxs': Var(data['stim']['stimIdxs'] - 1)})
        # stim.data has size nStim x nRuns.
        tstep = 1 / data['stim']['fs']
        stim_names = data['stim']['names']
        for name, stim_data in zip(stim_names, data['stim']['data']):
            stim_ndvars = []
            for run_data in stim_data:
                uts = UTS(0, tstep, len(run_data))
                stim_ndvars.append(NDVar(run_data, uts, name))
            key = Dataset.as_key(name)
            ds[key] = combine(stim_ndvars, to_list=True)
        # Condition labels
        if 'condIdxs' in data['stim']:
            ds['condIdxs'] = Var(data['stim']['condIdxs'] - 1)
            labels = dict(enumerate(data['stim']['condNames'], 1))
            ds['condition'] = Factor(data['stim']['condIdxs'], labels=labels)
        return ds
    raise IOError("File contains neither 'eeg' or 'stim' entry")



from eelbrain import datasets, plot
from eelbrain.testing import hide_plots


@hide_plots
def test_linestack():
    "Test plot.LineStack()"
    ds = datasets.get_uts()
    p = plot.LineStack(ds[:10, 'uts'])
    ax = p.figure.axes[0]
    assert ax.get_xlim() == (-0.2, .79)
    p.set_xlim(.1, .4)
    assert ax.get_xlim() == (.1, .4)
    p.close()



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""
.. _exa-mu-anova:

ANOVA
=====

.. currentmodule:: eelbrain

This example performs mass-univariate ANOVA using :class:`testnd.ANOVA`.
The example shows a 2 x 2 repeated measures ANOVA.
For specifying different ANOVA type models see the :class:`test.ANOVA` documentation.
The example uses simulated data meant to vaguely resemble data from an N400 experiment
(not intended as a physiologically realistic simulation).

.. contents:: Contents
   :local:

"""
from eelbrain import *

###############################################################################
# Simulated data
# --------------
# Use :func:`datasets.simulate_erp` to generate a dataset
# simulating an N400 experiment, as in the :ref:`exa-cluster-based-mu` example.
# Set ``short_long=True`` to simulate word length as a second variable,
# resulting in a 2 x 2, *predictability* x *word length* design.
s_datasets = []
for subject in range(10):
    # generate data for one subject
    s_data = datasets.simulate_erp(seed=subject, short_long=True)
    # average across trials to get condition means
    s_data_agg = s_data.aggregate('predictability % length')
    # add the subject name as variable
    s_data_agg[:, 'subject'] = f'S{subject:02}'
    s_datasets.append(s_data_agg)

data = combine(s_datasets)
# Define subject as random factor (to treat it as random effect in the ANOVA)
data['subject'].random = True
data.head()

###############################################################################
# Re-reference the EEG data (i.e., subtract the mean of the two mastoid channels):
data['eeg'] -= data['eeg'].mean(sensor=['M1', 'M2'])

###############################################################################
# Plot the data by condition to illustrate the effect. 
# Note 2 effects in the simulation: 
# an early ~130 ms peak determined by word `length`, simulating increased visual processing for longer words;
# and a later "N400" `predictability` peak, simulated to be larger for more surprising (less predictable) words.

p = plot.TopoButterfly('eeg', 'predictability % length', t=.400, data=data, axh=2, w=8, clip='circle')

###############################################################################
# Spatio-temporal test
# --------------------
# :class:`testnd.ANOVA` provides an interface for mass-univariate ANOVA.
# The ANOVA function determines whether to perform a repeated measures or fixed effects ANOVA based on the model.
# Here, ``'predictability * length * subject'`` is a repeated measures ANOVA because we defined ``data['subject']`` as random effect above.
# Changing the model to ``'predictability * length'`` would perform a fixed effects ANOVA.
result = testnd.ANOVA(
    'eeg', 
    'predictability * length * subject',
    data=data,
    pmin=0.05,  # Use uncorrected p = 0.05 as threshold for forming clusters
    tstart=0.050,  # Find clusters in the time window from 100 ...
    tstop=0.600,  # ... to 600 ms
    samples=1000,  # smaller number of permutations to speed up the example; use 10'000 when the exact p-value matters
)

###############################################################################
# The default visualization shows *F*-values over time.
# A yellow line indicates the F-value corresponding - the ``pmin=0.05`` parameter.
# In the butterfly plots, line segments that are part of a significant cluster are shown in color.
# In the topographic map, significant clusters are shown by outline.

p = plot.TopoButterfly(result, t=.400, axh=2, w=8, clip='circle', vmax=100)
_ = p.plot_colorbar()

###############################################################################
# The corresponding *F*-maps can be accessed on the `result` object:

result.f

###############################################################################
# Corresponding cluster statistics:

result.find_clusters(0.05)



"""
.. _exa-data_split:
.. currentmodule:: eelbrain

Data partitions for boosting
============================

The boosting algorithm can use two different forms of cross-validation: cross-validation as stopping criterion (always on), and cross-validation for model evaluation (optional). This requires paertitioning the data into different segments. The :func:`eelbrain.plot.preview_partitions` function is for exploring the effect of different parameters on the way the data is split.

.. contents:: Sections
    :local:
    :backlinks: top

Validation set
--------------
During boosting, every training step consists in modifying one element of the kernel/TRF. After every such step, the new TRF is evaluated against the validation data. For continuous data (without :class:`Case` dimension), the default is to split the data into 10 equal-length segments, and perform 10 model fits, each using one of the segments as validation set. In the plots below, each "Split" shown on the y-axis corresponds to a separate run of the boosting algorithm. The results returned by the :func:`boosting` function would be based on to the average TRF of those 10 runs.
"""
# sphinx_gallery_thumbnail_number = 6
from eelbrain import *


p = plot.preview_partitions()

###############################################################################
# The number of partitions can be controlled with the `partitions` parameter:

p = plot.preview_partitions(partitions=5)
p = plot.preview_partitions(partitions=7)

###############################################################################
# For data with multiple trials (data with a :class:`Case` dimension), the function attempts to use trials evenly across time: 

p = plot.preview_partitions(20, partitions=5)
p = plot.preview_partitions(20, partitions=2)

###############################################################################
# Validation and testing sets
# ---------------------------
# Testing the result of a model fit with cross-validation requires data that was never used during training. Testing with cross-validation is enabled in the :func:`boosting` function by setting ``test=True``. When testing is enabled, each data segment is used in turn as testing segment. For each testing segment, the remaining segment are used in different runs as training and validation data. The results of those runs are then averaged to predict responses in the testing data. This nested loop means that the number of boosting runs can get large quickly when using many partitions, so the default is to use just four partitions:

p = plot.preview_partitions(test=True)
p = plot.preview_partitions(20, partitions=5, test=True)



from surfer import Brain

DST = '../images/brain_%s.png'
subjects_dir = '/Volumes/Seagate/refpred/mri'
surfaces = ('inflated',
            'inflated_avg',
            'inflated_pre',
            # 'orig',
            # 'pial',
            'smoothwm',
            'sphere',
            'white')


for surf in surfaces:
    brain = Brain('fsaverage', 'lh', surf, size=400, background='white',
                  subjects_dir=subjects_dir)
    brain.save_image(DST % surf, 'rgba', True)
